"""Collapse multiple config-sources and instantiate from them.

A lot of extra documentation on this is in dev-notes/config.rst.
"""

__all__ = ("CollapsedConfig", "ConfigManager",)

import weakref
from collections import defaultdict, deque, namedtuple
from itertools import chain

from snakeoil import klass, mappings
from snakeoil.compatibility import IGNORED_EXCEPTIONS

from . import basics, errors

_section_data = namedtuple('_section_data', ['name', 'section'])


class _ConfigMapping(mappings.DictMixin):

    """Minimal dict-like wrapper returning config sections by type.

    Similar to :class:`mappings.LazyValDict` but __getitem__
    does not call the key func for __getitem__.

    Careful: getting the keys for this mapping will collapse all of
    central's configs to get at their types, which might be slow if
    any of them are remote!
    """

    def __init__(self, manager, typename):
        super().__init__()
        self.manager = manager
        self.typename = typename

    def __getitem__(self, key):
        conf = self.manager.collapse_named_section(key, raise_on_missing=False)
        if conf is None or conf.type.name != self.typename:
            raise KeyError(key)
        return conf.instantiate()

    def keys(self):
        for name in self.manager.sections():
            try:
                collapsed = self.manager.collapse_named_section(name)
            except errors.ConfigError:
                # Cannot be collapsed, ignore it (this is not
                # an error, it can be used as base for
                # something that can be collapsed)
                pass
            else:
                if collapsed.type.name == self.typename:
                    yield name

    def __contains__(self, key):
        conf = self.manager.collapse_named_section(key, raise_on_missing=False)
        return conf is not None and conf.type.name == self.typename


class _ConfigStack(defaultdict):

    def __init__(self):
        super().__init__(list)

    def render_vals(self, manager, key, type_name):
        for data in self.get(key, ()):
            if key in data.section:
                yield data.section.render_value(manager, key, type_name)

    def render_val(self, manager, key, type_name):
        for val in self.render_vals(manager, key, type_name):
            return val
        return None

    def render_prepends(self, manager, key, type_name, flatten=True):
        results = []
        # keep in mind that the sequence we get is a top -> bottom walk of the config
        # as such for this operation we have to reverse it when building the content-
        # specifically, reverse the ordering, but not the content of each item.
        data = []
        for content in self.render_vals(manager, key, type_name):
            data.append(content)
            if content[1]:
                break

        for prepend, this_content, append in reversed(data):
            if this_content:
                results = [this_content]
            if prepend:
                results = [prepend] + results
            if append:
                results += [append]

        if flatten:
            results = chain.from_iterable(results)
        return list(results)


class CollapsedConfig:

    """A collapsed config section.

    :type type: :obj:`basics.ConfigType`
    :ivar type: Our type.
    :type config: dict
    :ivar config: The supplied configuration values.
    :ivar debug: if True exception wrapping is disabled.
    :ivar default: True if this section is a default.
    :type name: C{str} or C{None}
    :ivar name: our section name or C{None} for an anonymous section.
    """

    def __init__(self, type_obj, config, manager, debug=False, default=False):
        """Initialize instance vars."""
        # Check if we got all values required to instantiate.
        missing = set(type_obj.required) - set(config)
        if missing:
            module = type_obj.callable.__module__
            name = type_obj.callable.__name__
            missing_vars = ', '.join(map(repr, missing))
            raise errors.ConfigurationError(
                f'type {module}.{name} needs settings for {missing_vars}')

        self.name = None
        self.default = default
        self.debug = debug
        self.type = type_obj
        self.config = config
        # Cached instance if we have one.
        self._instance = None
        if manager is not None:
            manager = weakref.ref(manager)
        self.manager = manager

    def instantiate(self):
        if self._instance is None:
            try:
                self._instance = self._instantiate()
            except IGNORED_EXCEPTIONS:
                raise
            except Exception as e:
                raise errors.InstantiationError(self.name) from e
        return self._instance

    def _instantiate(self):
        """Call our type's callable, cache and return the result.

        Calling instantiate more than once will return the cached value.
        """

        # Needed because this code can run twice even with instance
        # caching if we trigger an ComplexInstantiationError.
        config = mappings.ProtectedDict(self.config)

        # Instantiate section refs.
        # Careful: not everything we have for needs to be in the conf dict
        # (because of default values) and not everything in the conf dict
        # needs to have a type (because of allow_unknowns).
        for name, val in config.items():
            typename = self.type.types.get(name)
            if typename is None:
                continue
            # central already checked the type, no need to repeat that here.
            unlist_it = False
            if typename.startswith('ref:'):
                val = [val]
                unlist_it = True
            if typename.startswith('refs:') or unlist_it:
                try:
                    final_val = []
                    for ref in val:
                        final_val.append(ref.instantiate())
                except IGNORED_EXCEPTIONS:
                    raise
                except Exception as e:
                    raise errors.ConfigurationError(
                        f'Instantiating reference {name!r} pointing at {ref.name!r}') from e
                if unlist_it:
                    final_val = final_val[0]
                config[name] = final_val


        if self.type.requires_config:
            if self.manager is None:
                raise Exception(
                    'configuration internal error; '
                    'requires_config is enabled '
                    'but we have no config manager to return '
                )
            manager = self.manager()
            if manager is None:
                raise Exception(
                    'Configuration internal error, potentially '
                    'client code error; manager requested, but the config '
                    'manager is no longer in memory'
                )

            config[self.type.requires_config] = manager

        callable_obj = self.type.callable
        # return raw, uninstantiated class object if requested
        if self.type.raw_class:
            return callable_obj

        pargs = []
        for var in self.type.positional:
            pargs.append(config.pop(var))
        # Python is basically the worst language ever:
        # TypeError: repo() argument after ** must be a dictionary
        configdict = dict(config)
        try:
            self._instance = callable_obj(*pargs, **configdict)
        except IGNORED_EXCEPTIONS:
            raise
        except Exception as e:
            source = errors._identify_functor_source(self.type.callable)
            raise errors.InstantiationError(
                self.name, f'exception caught from {source!r}') from e
        if self._instance is None:
            raise errors.ComplexInstantiationError(
                'No object returned', callable_obj=callable_obj, pargs=pargs,
                kwargs=configdict)

        return self._instance

    def __getstate__(self):
        d = self.__dict__.copy()
        # pull actual value from weakref
        d['manager'] = d['manager']()
        return d

    def __setstate__(self, state):
        self.__dict__ = state.copy()
        # reset weakref
        self.__dict__['manager'] = weakref.ref(self.__dict__['manager'])


class _ConfigObjMap:

    def __init__(self, manager):
        self._manager = manager

    def __getattr__(self, attr):
        return _ConfigMapping(self._manager, attr)

    def __getitem__(self, key):
        val = getattr(self._manager.objects, key, klass.sentinel)
        if val is None:
            raise KeyError(key)
        return val

    def __getstate__(self):
        # Explicitly defined to force pickling to work as expected without
        # trying to pull __getstate__ from _ConfigMapping due to __getattr__.
        return self.__dict__.copy()

    def __setstate__(self, state):
        self.__dict__.update(state)


class CompatConfigManager:

    def __init__(self, manager):
        self._manager = manager

    def __getattr__(self, attr):
        if attr == '_manager':
            return object.__getattribute__(self, '_manager')
        obj = getattr(self._manager, attr, klass.sentinel)
        if obj is klass.sentinel:
            obj = getattr(self._manager.objects, attr)
        return obj

    __dir__ = klass.DirProxy("_manager")


class ConfigManager:

    """Combine config type definitions and configuration sections.

    Creates instances of a requested type and name by pulling the
    required data from any number of provided configuration sources.

    The following special type names are recognized:
      - configsection: instantiated and used the same way as an entry in the
        configs :obj:`__init__` arg.

    These "magic" typenames are only recognized if they are used by a
    section with a name starting with "autoload".
    """

    def __init__(self, configs=(), debug=False):
        """Initialize.

        :type configs: sequence of mappings of string to ConfigSection.
        :param configs: configuration to use.
            Can define extra configs that are also loaded.
        :param debug: if set to True exception wrapping is disabled.
            This means things can raise other exceptions than
            ConfigurationError but tracebacks are complete.
        """
        self.original_config_sources = tuple(map(self._compat_mangle_config, configs))
        # Set of encountered section names, used to catch recursive references.
        self._refs = set()
        self.debug = debug
        self.reload()
        # cycle...
        self.objects = _ConfigObjMap(self)

    def _compat_mangle_config(self, config):
        if hasattr(config, 'sections'):
            return config
        return basics.GeneratedConfigSource(config, "unknown")

    def reload(self):
        """Reinitialize from the config sources originally passed in.

        This throws away all cached instances and re-executes autoloads.
        """
        # "Attribute defined outside __init__"
        # pylint: disable-msg=W0201
        self.configs = []
        self.config_sources = []
        # Cache mapping confname to CollapsedConfig.
        self.rendered_sections = {}
        self.sections_lookup = defaultdict(deque)
        # force regeneration.
        self._types = klass._uncached_singleton
        for config in self.original_config_sources:
            self.add_config_source(config)

    def update(self, config):
        """Reinitialize using an additional supplied config."""
        self.original_config_sources += (config,)
        self.reload()

    def add_config_source(self, config):
        return self._add_config_source(self._compat_mangle_config(config))

    def _add_config_source(self, config):
        """Pull extra type and config sections from configs and use them.

        Things loaded this way are added after already loaded things
        (meaning the config containing the autoload section overrides
        the config(s) added by that section).
        """
        config_data = config.sections()

        collision = set(self.rendered_sections)
        collision.intersection_update(config_data)

        if collision:
            # If this matches something we previously instantiated
            # we should probably blow up to prevent massive
            # amounts of confusion (and recursive autoloads)
            sections = ', '.join(repr(x) for x in sorted(collision))
            raise errors.ConfigurationError(
                'New config is trying to modify existing section(s) '
                f'{sections} that was already instantiated.'
            )

        self.configs.append(config_data)
        self.config_sources.append(config)
        for name in config_data:
            self.sections_lookup[name].appendleft(config_data[name])

            # Do not even touch the ConfigSection if it's not an autoload.
            if not name.startswith('autoload'):
                continue

            try:
                collapsed = self.collapse_named_section(name)
            except IGNORED_EXCEPTIONS:
                raise
            except Exception as e:
                raise errors.ConfigurationError(
                    f'Failed collapsing autoload section {name!r}') from e

            if collapsed.type.name != 'configsection':
                raise errors.ConfigurationError(
                   f'Section {name!r} is marked as autoload but '
                   f'type is {collapsed.type.name}, not configsection'
                )
            try:
                instance = collapsed.instantiate()
            except IGNORED_EXCEPTIONS:
                raise
            except Exception as e:
                raise errors.AutoloadInstantiationError(name) from e
            if collapsed.type.name == 'configsection':
                self.add_config_source(instance)

    def sections(self):
        """Return an iterator of all section names."""
        return iter(self.sections_lookup.keys())

    def collapse_named_section(self, name, raise_on_missing=True):
        """Collapse a config by name, possibly returning a cached instance.

        @returns: :obj:`CollapsedConfig`.

        If there is no section with this name a ConfigurationError is raised,
        unless raise_on_missing is False in which case None is returned.
        """
        if name in self._refs:
            raise errors.ConfigurationError(f'Reference to {name!r} is recursive')
        self._refs.add(name)
        try:
            result = self.rendered_sections.get(name)
            if result is not None:
                return result
            section_stack = self.sections_lookup.get(name)
            if section_stack is None:
                if not raise_on_missing:
                    return None
                raise errors.ConfigurationError(f'no section called {name!r}')
            try:
                result = self.collapse_section(section_stack, name)
                result.name = name
            except IGNORED_EXCEPTIONS:
                raise
            except Exception as e:
                raise errors.ConfigurationError(
                    f'Collapsing section named {name!r}') from e
            self.rendered_sections[name] = result
            return result
        finally:
            self._refs.remove(name)

    def _get_inherited_sections(self, name, sections):
        # List of (name, ConfigSection, index) tuples, most specific first.
        slist = [(name, sections)]

        # first map out inherits.
        inherit_names = set([name])
        for current_section, section_stack in slist:
            current_conf = section_stack[0]
            if 'inherit' not in current_conf:
                continue
            prepend, inherits, append = current_conf.render_value(
                self, 'inherit', 'list')
            if prepend is not None or append is not None:
                raise errors.ConfigurationError(
                    'Prepending or appending to the inherit list makes no sense')
            for inherit in inherits:
                if inherit == current_section:
                    # self-inherit.  Mkae use of section_stack to handle this.
                    if len(section_stack) == 1:
                        # nothing else to self inherit.
                        raise errors.ConfigurationError(
                            f'Self-inherit {inherit!r} cannot be found')
                    if isinstance(section_stack, deque):
                        slist.append((inherit, list(section_stack)[1:]))
                    else:
                        slist.append((inherit, section_stack[1:]))
                else:
                    if inherit in inherit_names:
                        raise errors.ConfigurationError(
                            f'Inherit {inherit!r} is recursive')
                    inherit_names.add(inherit)
                    target = self.sections_lookup.get(inherit)
                    if target is None:
                        raise errors.ConfigurationError(
                            f'Inherit target {inherit!r} cannot be found')
                    slist.append((inherit, target))
        return [_section_data(name, stack[0]) for (name, stack) in slist]

    def _section_is_inherit_only(self, section):
        if 'inherit-only' in section:
            if section.render_value(self, 'inherit-only', 'bool'):
                return True
        return False

    def collapse_section(self, sections, _name=None):
        """Collapse a ConfigSection to a :obj:`CollapsedConfig`."""

        if self._section_is_inherit_only(sections[0]):
            if sections[0].render_value(self, 'inherit-only', 'bool'):
                raise errors.CollapseInheritOnly(
                    'cannot collapse inherit-only section')

        relevant_sections = self._get_inherited_sections(_name, sections)

        config_stack = _ConfigStack()
        for data in relevant_sections:
            for key in data.section.keys():
                config_stack[key].append(data)

        kls = config_stack.render_val(self, 'class', 'callable')
        if kls is None:
            raise errors.ConfigurationError('no class specified')
        type_obj = basics.ConfigType(kls)
        is_default = bool(config_stack.render_val(self, 'default', 'bool'))

        for key in ('inherit', 'inherit-only', 'class', 'default'):
            config_stack.pop(key, None)

        collapsed = CollapsedConfig(type_obj, self._render_config_stack(type_obj, config_stack),
            self, default=is_default, debug=self.debug)
        return collapsed

    @klass.jit_attr
    def types(self):
        type_map = defaultdict(dict)
        for name, sections in self.sections_lookup.items():
            if self._section_is_inherit_only(sections[0]):
                continue
            obj = self.collapse_named_section(name)
            type_map[obj.type.name][name] = obj
        return mappings.ImmutableDict(
            (k, mappings.ImmutableDict(v))
            for k,v in type_map.items())

    def _render_config_stack(self, type_obj, config_stack):
        conf = {}
        for key in config_stack:
            typename = type_obj.types.get(key)
            if typename is None:
                if not type_obj.allow_unknowns:
                    raise errors.ConfigurationError(f'Type of {key!r} unknown')
                typename = 'str'

            is_ref = typename.startswith('ref:')
            is_refs = typename.startswith('refs:')

            if typename.startswith('lazy_'):
                typename = typename[5:]

            if typename.startswith('refs:') or typename in ('list', 'str'):
                result = config_stack.render_prepends(self, key, typename, flatten=(typename != 'str'))
                if typename == 'str':
                    result = ' '.join(result)
            else:
                result = config_stack.render_val(self, key, typename)

            if is_ref:
                result = [result]
                is_refs = True

            if is_refs:
                try:
                    result = [ref.collapse() for ref in result]
                except IGNORED_EXCEPTIONS:
                    raise
                except Exception as e:
                    raise errors.ConfigurationError(
                        f'Failed collapsing section key {key!r}') from e
            if is_ref:
                result = result[0]

            conf[key] = result

        # Check if we got all values required to instantiate.
        missing = set(type_obj.required) - set(conf)
        if missing:
            module = type_obj.callable.__module__
            name = type_obj.callable.__name__
            missing_vars = ', '.join(map(repr, missing))
            raise errors.ConfigurationError(
                f'type {module}.{name} needs settings for {missing_vars}')

        return mappings.ImmutableDict(conf)

    def get_default(self, type_name):
        """Finds the configuration specified default obj of type_name.

        Returns C{None} if no defaults.
        """
        try:
            defaults = self.types.get(type_name, {}).items()
        except IGNORED_EXCEPTIONS:
            raise
        except Exception as e:
            raise errors.ConfigurationError(
                f'Collapsing defaults for {type_name!r}') from e
        defaults = [(name, section) for name, section in defaults if section.default]

        if not defaults:
            return None

        if len(defaults) > 1:
            defaults = ', '.join(map(repr, sorted(x[0] for x in defaults)))
            raise errors.ConfigurationError(
                f'type {type_name} incorrectly has multiple default sections: {defaults}')

        try:
            return defaults[0][1].instantiate()
        except IGNORED_EXCEPTIONS:
            raise
        except Exception as e:
            raise errors.ConfigurationError(
                f'failed instantiating default {type_name} {defaults[0][0]!r}') from e
        return None



"""
exceptions thrown by repository classes.

Need to extend the usage a bit further still.
"""

from ..exceptions import PkgcoreException, PkgcoreUserException


class RepoError(PkgcoreException):
    """General repository error."""

    def __init__(self, msg):
        self.msg = msg


class InitializationError(RepoError, PkgcoreUserException):
    """General repository initialization failure."""

    def __str__(self):
        return f"repo init failed: {self.msg}"


class InvalidRepo(InitializationError, PkgcoreUserException):
    """Repository is not a repo or is otherwise invalid."""

    def __str__(self):
        return f"invalid repo: {self.msg}"


class UnsupportedRepo(RepoError, PkgcoreUserException):
    """Repository uses an unknown EAPI or is otherwise not supported."""

    def __init__(self, repo):
        self.repo = repo

    def __str__(self):
        return (
            f'{self.repo.repo_id!r} repo: '
            f'unsupported repo EAPI {str(self.repo.eapi)!r}'
        )



from functools import partial

import pytest


def pytest_addoption(parser):
    parser.addoption(
        '--network', action='store_true', dest="network",
        default=False, help="allow network related tests to run")


def mark_network(config, func):
    """Decorator to add a 'net' mark and skip the test unless --network is passed."""
    skip_func = pytest.mark.skipif(
        not config.option.network,
        reason="needs --network option to run")
    return skip_func(pytest.mark.net(func))


def pytest_configure(config):
    pytest.mark_network = partial(mark_network, config)



"""
package wrapper class to override a packages attributes
"""

__all__ = ("MutatedPkg",)

from .base import wrapper


class MutatedPkg(wrapper):
    __slots__ = ("_overrides",)

    def __init__(self, pkg, overrides):
        """
        :param pkg: :obj:`pkgcore.package.metadata.package` to wrap
        :param overrides: is an attr -> instance mapping to substitute when
            the attr is requested
        """
        super().__init__(pkg)
        object.__setattr__(self, "_overrides", overrides)

    def __getattr__(self, attr):
        o = self._overrides.get(attr)
        if o is not None:
            return o
        return getattr(self._raw_pkg, attr)

    def __repr__(self):
        return '<%s pkg=%r overrides=%r @%#8x>' % (
            self.__class__.__name__, self._raw_pkg, tuple(self._overrides),
            id(self))

    def __str__(self):
        return '%s(%s, overrides=%s)' % \
            (self.__class__.__name__, self._raw_pkg, tuple(self._overrides))



from functools import partial
from pickle import dumps, loads

from pkgcore import test
from pkgcore.ebuild import atom, errors, restricts
from pkgcore.ebuild.cpv import CPV
from pkgcore.restrictions.boolean import AndRestriction
from pkgcore.test.misc import FakePkg, FakeRepo
from snakeoil.compatibility import cmp


class Test_atom(test.TestRestriction):

    class kls(atom.atom):
        __inst_caching__ = True
        __slots__ = ()

    kls = staticmethod(kls)

    def test_removed_features(self):
        # Ensure multi-slots no longer are allowed.
        self.assertRaises(errors.MalformedAtom, self.kls, "dev-util/diffball:1,2")
        self.assertEqual(self.kls("dev-util/diffball:1").slot, "1")

    def test_solutions(self):
        d = self.kls("=dev-util/diffball-0.7.1:2")
        self.assertEqual(list(d.iter_dnf_solutions()), [[d]])
        self.assertEqual(d.dnf_solutions(), [[d]])
        self.assertEqual(list(d.iter_cnf_solutions()), [[d]])
        self.assertEqual(d.cnf_solutions(), [[d]])
        bd = AndRestriction(*d.restrictions)
        self.assertEqual(list(d.iter_dnf_solutions(True)), bd.dnf_solutions())
        self.assertEqual(list(d.iter_cnf_solutions(True)), bd.cnf_solutions())
        self.assertEqual(d.dnf_solutions(True), bd.dnf_solutions())
        self.assertEqual(d.cnf_solutions(True), bd.cnf_solutions())

    def test_str_hash(self):
        for s in (
                "dev-util/diffball", "=dev-util/diffball-0.7.1",
                ">foon/bar-1:2[-4,3]", "=foon/bar-2*", "~foon/bar-2.3",
                "cat/pkg:0", "cat/pkg:5", "cat/pkg:0/5", "cat/pkg:5/5",
                "cat/pkg:=", "cat/pkg:0=", "cat/pkg:*",
                "!dev-util/diffball", "!=dev-util/diffball-0.7*",
                "foon/bar::gentoo", ">=foon/bar-10_alpha1:1::gentoo[-not,use]",
                "!!dev-util/diffball[use]"):
            self.assertEqual(str(self.kls(s)), s)
            self.assertEqual(hash(self.kls(s, disable_inst_caching=True)),
                hash(self.kls(s, disable_inst_caching=True)))

    def test_blockers(self):
        self.assertRaises(errors.MalformedAtom, self.kls,
            "!!dev-util/diffball", eapi='0')
        self.assertRaises(errors.MalformedAtom, self.kls,
            "!!dev-util/diffball", eapi='1')
        self.assertRaises(errors.MalformedAtom, self.kls,
            "!!!dev-util/diffball", eapi='2')
        for x in range(0, 2):
            obj = self.kls("!dev-util/diffball", eapi=str(x))
            self.assertTrue(obj.blocks)
            self.assertTrue(obj.blocks_temp_ignorable)
            self.assertFalse(obj.blocks_strongly)
        obj = self.kls("!!dev-util/diffball", eapi='2')
        self.assertTrue(obj.blocks)
        self.assertFalse(obj.blocks_temp_ignorable)
        self.assertTrue(obj.blocks_strongly)


    def test_iter(self):
        d = self.kls("!>=dev-util/diffball-0.7:1::gentoo[use,x]")
        self.assertEqual(list(d), list(d.restrictions))

    def test_pickling(self):
        a = self.kls("dev-util/diffball")
        self.assertEqual(a, loads(dumps(a)))
        a = self.kls("dev-util/diffball", negate_vers=True)
        self.assertEqual(a, loads(dumps(a)))

    def test_glob(self):
        self.assertRaises(errors.MalformedAtom, self.kls,
            "dev-util/diffball-1*")
        self.assertRaises(errors.MalformedAtom, self.kls,
            "dev-util/diffball-1.*")

        a = self.kls("=dev-util/diffball-1.2*")
        self.assertMatch(a, FakePkg("dev-util/diffball-1.2"))
        self.assertMatch(a, FakePkg("dev-util/diffball-1.2.0"))
        self.assertMatch(a, FakePkg("dev-util/diffball-1.2-r1"))
        self.assertMatch(a, FakePkg("dev-util/diffball-1.2_alpha"))
        self.assertNotMatch(a, FakePkg("dev-util/diffball-1"))

    def test_nonversioned(self):
        a = self.kls("kde-base/kde")
        self.assertMatch(a, CPV.unversioned("kde-base/kde"))
        self.assertNotMatch(a, CPV.unversioned("kde-base/kde2"))
        self.assertMatch(a, CPV.versioned("kde-base/kde-3"))

    def make_atom(self, s, ops, ver):
        l = []
        if -1 in ops:
            l.append(">")
        if 0 in ops:
            l.append("=")
        if 1 in ops:
            l.append("<")
        return self.kls(f"{''.join(l)}{s}-{ver}")

    def test_versioned(self):
        astr = "app-arch/tarsync"
        le_cpv = CPV.versioned(f"{astr}-0")
        eq_cpv = CPV.versioned(f"{astr}-1.1-r2")
        ge_cpv = CPV.versioned(f"{astr}-2")
        # <, =, >
        ops = (-1, 0, 1)

        for ops, ver in ((-1, "1.0"), (-1, "1.1"),
            (0, "1.1-r2"), (1, "1.1-r3"), (1, "1.2")):
            if not isinstance(ops, (list, tuple)):
                ops = (ops,)
            a = self.make_atom(astr, ops, ver)
            if -1 in ops:
                self.assertMatch(a, ge_cpv)
                self.assertMatch(a, eq_cpv)
                self.assertNotMatch(a, le_cpv)
            if 0 in ops:
                self.assertTrue(a.match(eq_cpv))
                if ops == (0,):
                    self.assertNotMatch(a, le_cpv)
                    self.assertNotMatch(a, ge_cpv)
            if 1 in ops:
                self.assertNotMatch(a, ge_cpv)
                self.assertMatch(a, eq_cpv)
                self.assertMatch(a, le_cpv)

    def test_norev(self):
        astr = "app-arch/tarsync"
        a = self.kls(f"~{astr}-1")
        self.assertMatch(a, CPV.versioned(f"{astr}-1"))
        self.assertMatch(a, CPV.versioned(f"{astr}-1-r1"))
        self.assertMatch(a, CPV.versioned(f"{astr}-1-r0"))
        self.assertNotMatch(a, CPV.versioned(f"{astr}-2"))
        self.assertRaises(errors.MalformedAtom, self.kls, "~{astr}-r1")
        self.assertRaises(errors.MalformedAtom, self.kls, "~{astr}-r2")
        # special case- yes -r0 effectively is None, but -r shouldn't be used
        # with ~
        self.assertRaises(errors.MalformedAtom, self.kls, "~{astr}-r0")

    def check_use(self, eapi, defaults=False):
        astr = "dev-util/bsdiff"
        c = FakePkg(f"{astr}-1", use=("debug",), iuse=("debug", "foon"), slot=1)

        kls = partial(self.kls, eapi=eapi)

        # Valid chars: [a-zA-Z0-9_@+-]
        kls(f'{astr}[zZaA09]')
        kls(f'{astr}[x@y]')
        kls(f'{astr}[x+y]')
        kls(f'{astr}[x-y]')
        kls(f'{astr}[x_y]')
        kls(f'{astr}[-x_y]')
        kls(f'{astr}[x?]')
        kls(f'{astr}[!x?]')
        kls(f'{astr}[x=]')
        kls(f'{astr}[!x=]')

        if defaults:
            kls(f'{astr}[x(+)]')
            kls(f'{astr}[x(-)]')
            self.assertRaises(errors.MalformedAtom, kls, f'{astr}[x(+-)]')
            self.assertRaises(errors.MalformedAtom, kls, f'{astr}[x(@)]')
            self.assertMatch(kls(f"{astr}[debug(+)]"), c)
            self.assertMatch(kls(f"{astr}[debug(-)]"), c)
            self.assertMatch(kls(f"{astr}[missing(+)]"), c)
            self.assertNotMatch(kls(f"{astr}[missing(-)]"), c)
            self.assertMatch(kls(f"{astr}[missing(+)]"), c)
            self.assertMatch(kls(f"{astr}[-missing(-)]"), c)
            self.assertNotMatch(kls(f"{astr}[-missing(+)]"), c)

            self.assertMatch(kls(f"{astr}[-missing(-),debug]"), c)
            self.assertNotMatch(kls(f"{astr}[-missing(+),debug(+)]"), c)
            self.assertMatch(kls(f"{astr}[missing(+),debug(+)]"), c)
        else:
            self.assertRaises(errors.MalformedAtom, kls, f'{astr}[x(+)]')
            self.assertRaises(errors.MalformedAtom, kls, f'{astr}[x(-)]')

        # '.' not a valid char in use deps
        self.assertRaises(errors.MalformedAtom, kls, f"{astr}[x.y]")

        # Use deps start with an alphanumeric char (non-transitive)
        self.assertRaises(errors.MalformedAtom, kls, f"{astr}[@x]")
        self.assertRaises(errors.MalformedAtom, kls, f"{astr}[_x]")
        self.assertRaises(errors.MalformedAtom, kls, f"{astr}[+x]")
        self.assertRaises(errors.MalformedAtom, kls, f"{astr}[-@x]")
        self.assertRaises(errors.MalformedAtom, kls, f"{astr}[-_x]")
        self.assertRaises(errors.MalformedAtom, kls, f"{astr}[-+x]")
        self.assertRaises(errors.MalformedAtom, kls, f"{astr}[--x]")

        self.assertMatch(kls(f"{astr}[debug]"), c)
        self.assertNotMatch(kls(f"{astr}[-debug]"), c)
        self.assertMatch(kls(f"{astr}[debug,-not]"), c)
        self.assertMatch(kls(f"{astr}:1[debug,-not]"), c)

        self.assertRaises(errors.MalformedAtom, kls, f"{astr}[]")
        self.assertRaises(errors.MalformedAtom, kls, f"{astr}[-]")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[foon")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[[fo]")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[x][y]")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[x]:1")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[x]a")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[--]")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[x??]")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[x=?]")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[x?=]")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[x==]")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[x??]")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[!=]")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[!?]")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[!!x?]")
        self.assertRaises(errors.MalformedAtom, kls, "dev-util/diffball[!-x?]")

    def test_slot(self):
        astr = "dev-util/confcache"
        c = FakePkg(f"{astr}-1", slot=1)
        self.assertNotMatch(self.kls(f"{astr}:0"), c)
        self.assertMatch(self.kls(f"{astr}:1"), c)
        self.assertNotMatch(self.kls(f"{astr}:2"), c)
        # note the above isn't compliant with eapi2/3; thus this test
        self.assertRaises(errors.MalformedAtom, self.kls, "dev-util/foo:0", eapi='0')

        # shouldn't puke, but has, thus checking"
        self.kls("sys-libs/db:4.4")
        self.kls(f"{astr}:azAZ.-+_09")
        self.kls(f"{astr}:_bar") # According to PMS, underscore and plus-sign are
        self.kls(f"{astr}:+bar") # not invalid first chars in a slot dep
        self.assertRaises(errors.MalformedAtom, self.kls, "dev-util/foo:")
        self.assertRaises(errors.MalformedAtom, self.kls, "dev-util/foo:1,,0")
        self.assertRaises(errors.MalformedAtom, self.kls, "dev-util/foo:1:")
        self.assertRaises(errors.MalformedAtom, self.kls, "dev-util/foo:-1")
        self.assertRaises(errors.MalformedAtom, self.kls, "dev-util/foo:.1")
        self.assertRaises(errors.MalformedAtom, self.kls, "dev-util/foo:1@2")

    def test_slot_operators_and_subslots(self):
        self.assertRaises(errors.MalformedAtom, self.kls, "sys-libs/db:*", eapi='4')
        self.kls("sys-libs/db:*", eapi='5')
        self.assertRaises(errors.MalformedAtom, self.kls, "sys-libs/db:=", eapi='4')
        self.kls("sys-libs/db:=", eapi='5')
        self.assertRaises(errors.MalformedAtom, self.kls, "sys-libs/db:==", eapi='5')
        self.assertRaises(errors.MalformedAtom, self.kls, "sys-libs/db:1=", eapi='4')
        self.assertRaises(errors.MalformedAtom, self.kls, "sys-libs/db:2/3.0=", eapi='4')
        self.assertRaises(errors.MalformedAtom, self.kls, "sys-libs/db:2/3.0", eapi='1')
        self.assertRaises(errors.MalformedAtom, self.kls, "sys-libs/db:/=", eapi='5')
        self.assertRaises(errors.MalformedAtom, self.kls, "sys-libs/db:/1=", eapi='5')
        self.assertRaises(errors.MalformedAtom, self.kls, "sys-libs/db:1/=", eapi='5')
        self.assertRaises(errors.MalformedAtom, self.kls, "sys-libs/db:*1/=", eapi='5')

        for subslot in ("/1.0", ""):
            self.assertRaises(errors.MalformedAtom, self.kls, f"sys-libs/db:*4{subslot}", eapi='5')
            self.assertRaises(errors.MalformedAtom, self.kls, f"sys-libs/db:4{subslot}*", eapi='5')
            self.assertRaises(errors.MalformedAtom, self.kls, f"sys-libs/db:=4{subslot}", eapi='5')
            self.kls(f"sys-libs/db:4{subslot}=", eapi='5')
            self.kls(f"sys-libs/db:3.2{subslot}=", eapi='5')
            self.assertRaises(errors.MalformedAtom, self.kls, f"sys-libs/db:4{subslot}==", eapi='5')

        def check_it(text, slot, subslot, operator):
            obj = self.kls(f"sys-libs/db{text}")
            self.assertEqual(obj.slot, slot)
            self.assertEqual(obj.subslot, subslot)
            self.assertEqual(obj.slot_operator, operator)
        check_it(":4", "4", None, None)
        check_it(":=", None, None, "=")
        check_it(":4=", "4", None, "=")
        check_it(":4/0.4=", "4", "0.4", "=")
        check_it(":*", None, None, "*")

        # Verify restrictions.
        self.assertMatch(self.kls("sys-libs/db:1="),
            FakePkg("sys-libs/db-1", slot="1"))
        self.assertMatch(self.kls("sys-libs/db:1/2="),
            FakePkg("sys-libs/db-1", slot="1", subslot="2"))
        self.assertNotMatch(self.kls("sys-libs/db:1/2.3="),
            FakePkg("sys-libs/db-1", slot="1", subslot="2"))
        self.assertNotMatch(self.kls("sys-libs/db:1/2.3="),
            FakePkg("sys-libs/db-1", slot="1"))
        self.assertMatch(self.kls("sys-libs/db:1a.2/2.3"),
            FakePkg("sys-libs/db-1", slot="1a.2", subslot="2.3"))

    def test_getattr(self):
        # assert it explodes for bad attr access.
        obj = self.kls("dev-util/diffball")
        self.assertRaises(AttributeError, getattr, obj, "__foasdfawe")
        # assert ordering


        def assertAttr(attr):
            self.assertEqual(restrictions[pos].attr, attr,
                msg="expected attr %r at %i for ver(%s), repo(%s) use(%s), "
                    "slot(%s): got %r from %r" % (attr, pos, ver, repo, use,
                    slot, restrictions[pos].attr, restrictions))
            return pos + 1

        slot = ''
        def f():
            for pref, ver in (('', ''), ('=', '-0.1')):
                for repo in ('', '::gentoo'):
                    for slot in ('', ':1'):
                        for use in ('', '[x]'):
                            yield pref, ver, repo, slot, use

        for pref, ver, repo, slot, use in f():
            pos = 0
            o = self.kls(f"{pref}dev-util/diffball{ver}{slot}{repo}{use}")
            count = 2
            for x in ("use", "repo", "pref", "slot"):
                if locals()[x]:
                    count += 1

            restrictions = o.restrictions
            self.assertEqual(len(restrictions), count,
                msg="%r, restrictions count must be %i, got %i" %
                    (o, count, len(restrictions)))
            self.assertTrue([getattr(x, 'type', None)
                for x in restrictions], ['package'] * count)
            if repo:
                pos = assertAttr('repo.repo_id')
            pos = assertAttr('package')
            pos = assertAttr('category')
            if ver:
                self.assertInstance(restrictions[pos], restricts.VersionMatch,
                    msg="expected %r, got %r; repo(%s), ver(%s), use(%s) "
                        "slot(%s)" %
                        (restricts.VersionMatch, restrictions[pos],
                            repo, ver, use, slot))
                pos += 1
            if slot:
                pos = assertAttr('slot')
            if use:
                pos = assertAttr('use')

    def test_eapi0(self):
        for postfix in (':1', ':1,2', ':asdf', '::asdf', '::asdf-x86', '[x]',
                        '[x,y]', ':1[x,y]', '[x,y]:1', ':1::repo'):
            self.assertRaisesMsg(f"dev-util/foon{postfix} must be invalid in EAPI 0",
                errors.MalformedAtom, self.kls, f"dev-util/foon{postfix}", eapi='0')

    def test_eapi1(self):
        for postfix in (':1,2', '::asdf', '::asdf-x86', '[x]',
                        '[x,y]', ':1[x,y]', '[x,y]:1', ':1:repo'):
            self.assertRaisesMsg(f"dev-util/foon{postfix} must be invalid in EAPI 1",
                errors.MalformedAtom, self.kls,
                f"dev-util/foon{postfix}", eapi='1')
        self.kls("dev-util/foon:1", eapi='1')
        self.kls("dev-util/foon:12", eapi='1')
        self.assertRaisesMsg("dev-util/foon[dar] must be invalid in EAPI 1",
            errors.MalformedAtom,
            self.kls, "dev-util/foon[dar]", eapi='1')

    def test_eapi2(self):
        self.check_use('2')

    def test_eapi3(self):
        self.check_use('3')
        self.kls("dev-util/foon:1", eapi='3')
        self.kls("dev-util/foon:2", eapi='3')
        self.kls("!dev-util/foon:1", eapi='3')
        self.kls("dev-util/foon:1[x]", eapi='3')
        self.kls("dev-util/foon:1[x?]", eapi='3')
        self.assertRaises(errors.MalformedAtom, self.kls, "dev-util/foon:1::dar", eapi='3')

    def test_eapi4(self):
        self.check_use('4', defaults=True)

    def test_eapi5(self):
        self.check_use('5', defaults=True)

    def test_eapi6(self):
        self.check_use('6', defaults=True)

    def test_eapi7(self):
        self.check_use('7', defaults=True)

    def test_repo_id(self):
        astr = "dev-util/bsdiff"
        c = FakePkg(f"{astr}-1", repo=FakeRepo(repo_id="gentoo-x86A_"), slot="0")
        self.assertMatch(self.kls(f"{astr}"), c)
        self.assertMatch(self.kls(f"{astr}::gentoo-x86A_"), c)
        self.assertMatch(self.kls(f"{astr}:0::gentoo-x86A_"), c)
        self.assertNotMatch(self.kls(f"{astr}::gentoo2"), c)
        self.assertRaises(errors.MalformedAtom, self.kls, "dev-util/foon:1:")
        self.assertRaises(errors.MalformedAtom, self.kls, "dev-util/foon::")
        self.assertRaises(errors.MalformedAtom, self.kls, "dev-util/foon::-gentoo-x86")
        self.assertRaises(errors.MalformedAtom, self.kls, "dev-util/foon:::")
        for x in range(0, 3):
            self.assertRaises(errors.MalformedAtom, self.kls,
                "dev-util/foon::gentoo-x86", eapi=str(x))

    def test_invalid_atom(self):
        self.assertRaises(errors.MalformedAtom, self.kls, '~dev-util/spork')
        self.assertRaises(errors.MalformedAtom, self.kls, '>dev-util/spork')
        self.assertRaises(errors.MalformedAtom, self.kls, 'dev-util/spork-3')
        self.assertRaises(errors.MalformedAtom, self.kls, 'spork')

    def test_intersects(self):
        for this, that, result in [
            ('cat/pkg', 'pkg/cat', False),
            ('cat/pkg', 'cat/pkg', True),
            ('cat/pkg:1', 'cat/pkg:2', False),
            ('cat/pkg:1', 'cat/pkg:1', True),
            ('cat/pkg:1', 'cat/pkg[foo]', True),
            ('cat/pkg:0/0', 'cat/pkg:0/1', False),
            ('cat/pkg:0/0', 'cat/pkg:0/0', True),
            ('cat/pkg:0/0', 'cat/pkg:0', True),
            ('cat/pkg:0/0', 'cat/pkg', True),
            ('cat/pkg[foo]', 'cat/pkg[-bar]', True),
            ('cat/pkg[foo]', 'cat/pkg[-foo]', False),
            ('>cat/pkg-3', '>cat/pkg-1', True),
            ('>cat/pkg-3', '<cat/pkg-3', False),
            ('>=cat/pkg-3', '<cat/pkg-3', False),
            ('>cat/pkg-2', '=cat/pkg-2*', True),
            ('<cat/pkg-2_alpha1', '=cat/pkg-2*', True),
            ('=cat/pkg-2', '=cat/pkg-2', True),
            ('=cat/pkg-3', '=cat/pkg-2', False),
            ('=cat/pkg-2', '>cat/pkg-2', False),
            ('=cat/pkg-2', '>=cat/pkg-2', True),
            ('~cat/pkg-2', '~cat/pkg-2', True),
            ('~cat/pkg-2', '~cat/pkg-2.1', False),
            ('=cat/pkg-2*', '=cat/pkg-2.3*', True),
            ('>cat/pkg-2.4', '=cat/pkg-2*', True),
            ('<cat/pkg-2.4', '=cat/pkg-2*', True),
            ('<cat/pkg-1', '=cat/pkg-2*', False),
            ('~cat/pkg-2', '>cat/pkg-2-r1', True),
            ('~cat/pkg-2', '<=cat/pkg-2', True),
            ('=cat/pkg-2-r2*', '<=cat/pkg-2-r20', True),
            ('=cat/pkg-2-r2*', '<cat/pkg-2-r20', True),
            ('=cat/pkg-2-r2*', '<=cat/pkg-2-r2', True),
            ('~cat/pkg-2', '<cat/pkg-2', False),
            ('=cat/pkg-1-r10*', '~cat/pkg-1', True),
            ('=cat/pkg-1-r1*', '<cat/pkg-1-r1', False),
            ('=cat/pkg-1*', '>cat/pkg-2', False),
            ('>=cat/pkg-8.4', '=cat/pkg-8.3.4*', False),
            ('cat/pkg::gentoo', 'cat/pkg', True),
            ('cat/pkg::gentoo', 'cat/pkg::foo', False),
            # known to cause an assplosion, thus redundant test.
            ('=sys-devel/gcc-4.1.1-r3', '=sys-devel/gcc-3.3*', False),
            ('=sys-libs/db-4*', '~sys-libs/db-4.3.29', True),
            ]:
            this_atom = self.kls(this)
            that_atom = self.kls(that)
            self.assertEqual(
                result, this_atom.intersects(that_atom),
                f'{this} intersecting {that} should be {result}')
            self.assertEqual(
                result, that_atom.intersects(this_atom),
                f'{that} intersecting {this} should be {result}')

    def assertEqual2(self, o1, o2):
        # logic bugs hidden behind short circuiting comparisons for metadata
        # is why we test the comparison *both* ways.
        self.assertEqual(o1, o2)
        c = cmp(o1, o2)
        self.assertEqual(c, 0,
            msg="checking cmp for %r, %r, aren't equal: got %i" % (o1, o2, c))
        self.assertEqual(o2, o1)
        c = cmp(o2, o1)
        self.assertEqual(c, 0,
            msg="checking cmp for %r, %r,aren't equal: got %i" % (o2, o1, c))

    def assertNotEqual2(self, o1, o2):
        # is why we test the comparison *both* ways.
        self.assertNotEqual(o1, o2)
        c = cmp(o1, o2)
        self.assertNotEqual(c, 0,
            msg="checking cmp for %r, %r, not supposed to be equal, got %i"
                % (o1, o2, c))
        self.assertNotEqual(o2, o1)
        c = cmp(o2, o1)
        self.assertNotEqual(c, 0,
            msg="checking cmp for %r, %r, not supposed to be equal, got %i"
                % (o2, o1, c))

    def test_comparison(self):
        self.assertEqual2(self.kls('cat/pkg'), self.kls('cat/pkg'))
        self.assertNotEqual2(self.kls('cat/pkg'), self.kls('cat/pkgb'))
        self.assertNotEqual2(self.kls('cata/pkg'), self.kls('cat/pkg'))
        self.assertNotEqual2(self.kls('cat/pkg'), self.kls('!cat/pkg'))
        self.assertEqual2(self.kls('!cat/pkg'), self.kls('!cat/pkg'))
        self.assertNotEqual2(self.kls('=cat/pkg-0.1:0'),
            self.kls('=cat/pkg-0.1'))
        self.assertNotEqual2(self.kls('=cat/pkg-1[foon]'),
            self.kls('=cat/pkg-1'))
        self.assertEqual2(self.kls('=cat/pkg-0'), self.kls('=cat/pkg-0'))
        self.assertNotEqual2(self.kls('<cat/pkg-2'), self.kls('>cat/pkg-2'))
        self.assertNotEqual2(self.kls('=cat/pkg-2*'), self.kls('=cat/pkg-2'))
        self.assertNotEqual2(self.kls('=cat/pkg-2', True),
            self.kls('=cat/pkg-2'))

        # use...
        self.assertNotEqual2(self.kls('cat/pkg[foo]'), self.kls('cat/pkg'))
        self.assertNotEqual2(self.kls('cat/pkg[foo]'),
                             self.kls('cat/pkg[-foo]'))
        self.assertEqual2(self.kls('cat/pkg[foo,-bar]'),
                          self.kls('cat/pkg[-bar,foo]'))
        # repo_id
        self.assertEqual2(self.kls('cat/pkg::a'), self.kls('cat/pkg::a'))
        self.assertNotEqual2(self.kls('cat/pkg::a'), self.kls('cat/pkg::b'))
        self.assertNotEqual2(self.kls('cat/pkg::a'), self.kls('cat/pkg'))

        # slots.
        self.assertNotEqual2(self.kls('cat/pkg:1'), self.kls('cat/pkg'))
        self.assertEqual2(self.kls('cat/pkg:2'), self.kls('cat/pkg:2'))
        for lesser, greater in (('0.1', '1'), ('1', '1-r1'), ('1.1', '1.2')):
            self.assertTrue(self.kls(f'=d/b-{lesser}') <
                self.kls(f'=d/b-{greater}'),
                msg=f"d/b-{lesser} < d/b-{greater}")
            self.assertFalse(self.kls(f'=d/b-{lesser}') >
                self.kls(f'=d/b-{greater}'),
                msg=f"!: d/b-{lesser} < d/b-{greater}")
            self.assertTrue(self.kls(f'=d/b-{greater}') >
                self.kls(f'=d/b-{lesser}'),
                msg=f"d/b-{greater} > d/b-{lesser}")
            self.assertFalse(self.kls(f'=d/b-{greater}') <
                self.kls(f'=d/b-{lesser}'),
                msg=f"!: d/b-{greater} > d/b-{lesser}")

        self.assertTrue(self.kls("!!=d/b-1", eapi='2') > self.kls("!=d/b-1"))
        self.assertTrue(self.kls("!=d/b-1") < self.kls("!!=d/b-1"))
        self.assertEqual(self.kls("!=d/b-1"), self.kls("!=d/b-1"))

    def test_compatibility(self):
        self.assertNotMatch(self.kls('=dev-util/diffball-0.7'),
            FakePkg('dev-util/diffball-0.7.0'))
        # see bug http://bugs.gentoo.org/152127
        self.assertNotMatch(self.kls('>=sys-apps/portage-2.1.0_pre3-r5'),
            FakePkg('sys-apps/portage-2.1_pre3-r5'))

    def test_combined(self):
        p = FakePkg('dev-util/diffball-0.7', repo=FakeRepo(repo_id='gentoo'))
        self.assertMatch(self.kls('=dev-util/diffball-0.7::gentoo'), p)
        self.assertMatch(self.kls('dev-util/diffball::gentoo'), p)
        self.assertNotMatch(self.kls('=dev-util/diffball-0.7:1::gentoo'),
            FakePkg('dev-util/diffball-0.7', slot='2'))

    def test_unversioned(self):
        self.assertTrue(self.kls("dev-util/diffball").is_simple)
        self.assertFalse(self.kls("dev-util/diffball:2").is_simple)
        self.assertFalse(self.kls("dev-util/diffball:2::gentoo").is_simple)
        self.assertFalse(self.kls("dev-util/diffball::gentoo").is_simple)
        self.assertFalse(self.kls("!=dev-util/diffball-1").is_simple)
        self.assertFalse(self.kls(">dev-util/diffball-1.2").is_simple)
        self.assertFalse(self.kls("=dev-util/diffball-1").is_simple)
        self.assertFalse(self.kls("dev-util/diffball[x]").is_simple)
        self.assertFalse(self.kls("dev-util/diffball[x?]").is_simple)



__all__ = ("bzr_syncer",)

import os

from snakeoil.process.spawn import spawn_get_output

from . import base


class bzr_syncer(base.VcsSyncer):

    binary = "bzr"

    supported_uris = (
        ('bzr+', 5),
    )

    @classmethod
    def is_usable_on_filepath(cls, path):
        bzr_path = os.path.join(path, '.bzr')
        if cls.disabled or not os.path.isdir(bzr_path):
            return None
        code, data = spawn_get_output([cls.binary, "info", path])
        if code != 0:
            # should alert the user somehow
            return None
        for line in data:
            line = line.strip().split(":", 1)
            if len(line) != 2:
                continue
            if line[0] == 'parent branch':
                uri = f"bzr+{line[1].strip()}"
                return (cls._rewrite_uri_from_stat(bzr_path, uri),)
        return None

    @staticmethod
    def parse_uri(raw_uri):
        if not raw_uri.startswith("bzr+"):
            raise base.UriError(raw_uri, "doesn't start with bzr+")
        return raw_uri[4:]

    def _initial_pull(self):
        return [self.binary_path, "branch", self.uri, self.basedir]

    def _update_existing(self):
        return [self.binary_path, "pull", self.uri]



__all__ = ("get_raw_pkg", "groupby_pkg")

import itertools
from operator import attrgetter


def get_raw_pkg(pkg):
    p = pkg
    while hasattr(p, "_raw_pkg"):
        p = p._raw_pkg
    return p


def groupby_pkg(iterable):
    for key, pkgs in itertools.groupby(iterable, attrgetter('key')):
        yield pkgs



from pkgcore.resolver import plan
from pkgcore.test.misc import FakePkg
from snakeoil.currying import post_curry
from snakeoil.test import TestCase


class TestPkgSorting(TestCase):

    def check_it(self, sorter, vers, expected, iter_sort_target=False):
        pkgs = [FakePkg(f"d-b/a-{x}") for x in vers]
        if iter_sort_target:
            pkgs = [[x, []] for x in pkgs]
        pkgs = list(sorter(pkgs))
        if iter_sort_target:
            pkgs = [x[0] for x in pkgs]
        self.assertEqual([int(x.fullver) for x in pkgs], expected)

    test_highest_iter_sort = post_curry(check_it, plan.highest_iter_sort,
        [7,9,3,2], [9,7,3,2], True)

    test_lowest_iter_sort = post_curry(check_it, plan.lowest_iter_sort,
        [7,9,4,2], [2,4,7,9], True)

    test_pkg_sort_highest = post_curry(check_it, plan.pkg_sort_highest,
        [1,9,7,10], [10,9,7,1])

    test_pkg_sort_lowest = post_curry(check_it, plan.pkg_sort_lowest,
        [11,9,1,6], [1,6,9,11])



"""tests for pkgcore.config's package __init__.py"""

import operator

from pkgcore.config import basics, load_config
from pkgcore.config.hint import configurable
from snakeoil.test import TestCase
from snakeoil.test.mixins import mk_named_tempfile


@configurable(typename='foo')
def passthrough(*args, **kwargs):
    return args, kwargs


class ConfigLoadingTest(TestCase):

    def setUp(self):
        self.user_config = mk_named_tempfile()
        self.user_config.write(
            '[foo]\n'
            'class = tests.config.test_init.passthrough\n'
            )
        self.user_config.flush()
        self.system_config = mk_named_tempfile()
        self.system_config.write(
            '[foo]\n'
            'class = also invalid\n'
            )
        self.system_config.flush()

    def tearDown(self):
        self.user_config.close()
        self.system_config.close()
        del self.user_config
        del self.system_config

    def test_load_config(self):
        manager = load_config(user_conf_file=self.user_config.name)
        self.assertEqual(manager.foo['foo'], ((), {}))

        # Test user config overrides system config.
        manager = load_config(
            user_conf_file=self.user_config.name,
            system_conf_file=self.system_config.name)
        self.assertEqual(manager.foo['foo'], ((), {}))

        # Test prepends.
        manager = load_config(
            user_conf_file=self.user_config.name,
            prepend_sources=[{'myfoo': basics.HardCodedConfigSection({
                            'inherit': ['foo']})}])
        self.assertEqual(manager.foo['myfoo'], ((), {}))

        # Test disabling loading.
        manager = load_config(
            user_conf_file=self.user_config.name,
            skip_config_files=True)
        self.assertRaises(
            KeyError,
            operator.getitem, manager.foo, 'foo')



import os
from unittest import mock

import pytest
from pkgcore.sync import base, git
from snakeoil.process import CommandNotFound


class TestGitSyncer:

    @pytest.fixture(autouse=True)
    def _setup(self, tmp_path):
        self.repo_path = tmp_path / 'repo'

    def test_uri_parse(self):
        assert git.git_syncer.parse_uri("git+http://dar") == "http://dar"

        with pytest.raises(base.UriError):
            git.git_syncer.parse_uri("git+://dar")

        # external binary doesn't exist
        with mock.patch('snakeoil.process.find_binary') as find_binary:
            find_binary.side_effect = CommandNotFound('git')
            with pytest.raises(base.SyncError):
                git.git_syncer(str(self.repo_path), "git+http://foon.com/dar")

        # fake that the external binary exists
        with mock.patch('snakeoil.process.find_binary') as find_binary:
            find_binary.return_value = 'git'
            for proto in ('http', 'https'):
                for uri in (f"git+{proto}://repo.git", f"{proto}://repo.git"):
                    o = git.git_syncer(str(self.repo_path), uri)
                    assert o.uri == f"{proto}://repo.git"

    @mock.patch('snakeoil.process.spawn.spawn')
    def test_sync(self, spawn):
        uri = 'git://foo.git'
        with mock.patch('snakeoil.process.find_binary', return_value='git'):
            syncer = git.git_syncer(str(self.repo_path), uri)
        # initial sync
        syncer.sync()
        assert spawn.call_args[0] == (
            ['git', 'clone', uri, str(self.repo_path) + os.path.sep],)
        assert spawn.call_args[1]['cwd'] is None
        # repo update
        self.repo_path.mkdir()
        syncer.sync()
        assert spawn.call_args[0] == (['git', 'pull'],)
        assert spawn.call_args[1]['cwd'] == syncer.basedir


@pytest.mark_network
class TestGitSyncerReal:

    def test_sync(self, tmp_path):
        path = tmp_path / 'repo'
        syncer = git.git_syncer(str(path), "https://github.com/pkgcore/pkgrepo.git")
        assert syncer.sync()
        assert os.path.exists(os.path.join(path, 'metadata', 'layout.conf'))
        assert syncer.sync()



"""
base restriction class
"""

from functools import partial

from snakeoil import caching, klass
from snakeoil.currying import pretty_docs


class base(klass.SlotsPicklingMixin, metaclass=caching.WeakInstMeta):
    """base restriction matching object.

    all derivatives *should* be __slots__ based (lot of instances may
    wind up in memory).
    """
    __inst_caching__ = True

    # __weakref__ here is implicit via the metaclass
    __slots__ = ()
    package_matching = False

    klass.inject_immutable_instance(locals())

    def match(self, *arg, **kwargs):
        raise NotImplementedError

    def force_False(self, *arg, **kwargs):
        return not self.match(*arg, **kwargs)

    def force_True(self, *arg, **kwargs):
        return self.match(*arg, **kwargs)

    def __len__(self):
        return 1


class AlwaysBool(base):
    """restriction that always yields a specific boolean"""

    __slots__ = ("type", "negate")

    __inst_caching__ = True

    def __init__(self, node_type=None, negate=False):
        """
        :param node_type: the restriction type the instance should be,
            typically :obj:`pkgcore.restrictions.restriction.package_type` or
            :obj:`pkgcore.restrictions.restriction.value_type`
        :param negate: boolean to return for the match
        """
        object.__setattr__(self, "negate", negate)
        object.__setattr__(self, "type", node_type)

    def match(self, *a, **kw):
        return self.negate

    def force_True(self, *a, **kw):
        return self.negate

    def force_False(self, *a, **kw):
        return not self.negate

    def __iter__(self):
        return iter(())

    def __str__(self):
        return f"always '{self.negate}'"

    def __repr__(self):
        return '<%s always %r @%#8x>' % (
            self.__class__.__name__, self.negate, id(self))

    def __getstate__(self):
        return self.negate, self.type

    def __setstate__(self, state):
        negate, node_type = state
        object.__setattr__(self, "negate", negate)
        object.__setattr__(self, "type", node_type)


class Negate(base):
    """wrap and negate a restriction instance"""

    __slots__ = ("type", "_restrict")
    __inst_caching__ = False

    def __init__(self, restrict):
        """
        :param restrict: :obj:`pkgcore.restrictions.restriction.base` instance
            to negate
        """
        sf = object.__setattr__
        sf(self, "type", restrict.type)
        sf(self, "_restrict", restrict)

    def match(self, *a, **kw):
        return not self._restrict.match(*a, **kw)

    def __str__(self):
        return "not (%s)" % self._restrict


class FakeType(base):
    """wrapper to wrap and fake a node_type"""

    __slots__ = ("type", "_restrict")
    __inst_caching__ = False

    def __init__(self, restrict, new_type):
        """
        :param restrict: :obj:`pkgcore.restrictions.restriction.base` instance
            to wrap
        :param new_type: new node_type
        """
        sf = object.__setattr__
        sf(self, "type", new_type)
        sf(self, "_restrict", restrict)

    def match(self, *a, **kw):
        return self._restrict.match(*a, **kw)

    def __str__(self):
        return "Faked type(%s): %s" % (self.type, self._restrict)


class AnyMatch(base):
    """Apply a nested restriction to every item in a sequence."""

    __slots__ = ('restriction', 'type', 'negate')

    def __init__(self, childrestriction, node_type, negate=False):
        """Initialize.

        :type childrestriction: restriction
        :param childrestriction: child restriction applied to every value.
        :type node_type: string
        :param node_type: type of this restriction.
        """
        sf = object.__setattr__
        sf(self, "negate", negate)
        sf(self, "restriction", childrestriction)
        sf(self, "type", node_type)

    def match(self, val):
        for x in val:
            if self.restriction.match(x):
                return not self.negate
        return self.negate

    def __str__(self):
        return "any: %s match" % (self.restriction,)

    def __repr__(self):
        return '<%s restriction=%r @%#8x>' % (
            self.__class__.__name__, self.restriction, id(self))


def curry_node_type(cls, node_type, extradoc=None):
    """Helper function for creating restrictions of a certain type.

    This uses :obj:`partial` to pass a node_type to the wrapped class,
    and extends the docstring.

    :param cls: callable (usually a class) that is wrapped.
    :param node_type: value passed as node_type.
    :param extradoc: addition to the docstring. Defaults to
        "Automatically set to %s type." % node_type

    :return: a wrapped callable.
    """
    if extradoc is None:
        extradoc = "Automatically set to %s type." % (node_type,)
    doc = cls.__doc__
    result = partial(cls, node_type=node_type)
    if doc is None:
        doc = ''
    else:
        # do this so indentation on pydoc __doc__ is sane
        doc = "\n".join(line.lstrip() for line in doc.split("\n")) + "\n"
        doc += extradoc
    return pretty_docs(result, doc)


value_type = "values"
package_type = "package"
valid_types = (value_type, package_type)



#!/usr/bin/env python3
# Copyright 2007 Charlie Shepherd

import sys

try:
    from pkgcore.repository.util import get_raw_repos, get_virtual_repos
    from pkgcore.util import commandline
except ImportError:
    print('Cannot import pkgcore!', file=sys.stderr)
    print('Verify it is properly installed and/or PYTHONPATH is set correctly.', file=sys.stderr)
    if '--debug' not in sys.argv:
        print('Add --debug to the commandline for a traceback.', file=sys.stderr)
    else:
        raise
    sys.exit(1)


argparser = commandline.ArgumentParser(color=False, version=False)
argparser.add_argument(
    '-r', '--repo', action=commandline.StoreRepoObject,
    help='repo to give info about (default from domain if omitted)')


@argparser.bind_final_check
def check_args(parser, namespace):
    # Get repo(s) to operate on.
    if namespace.repo:
        repos = (namespace.repo,)
    else:
        repos = namespace.domain.repos
    namespace.repos = get_virtual_repos(get_raw_repos(repos), False)


@argparser.bind_main_func
def main(options, out, err):
    for repo in options.repos:
        out.write("Repo ID: %s" % repo.repo_id)
        location = getattr(repo, "location", None)
        if location:
            out.write("Repo location: %s" % location)
        else:
            out.write("Repo has no on-disk location")
        out.write("%d packages" % len(repo.versions))
        out.write("%d categories" % len(repo.packages))
        out.write()


if __name__ == '__main__':
    tool = commandline.Tool(argparser)
    sys.exit(tool())



"""
restriction related utilities
"""

from snakeoil.sequences import iflatten_func

from . import boolean, packages, restriction


def _is_package_instance(inst):
    return (getattr(inst, "type", None) == restriction.package_type and not
            isinstance(inst, boolean.base))


def collect_package_restrictions(restrict, attrs=None, invert=False):
    """Collect PackageRestriction instances inside a restriction.

    :param restrict: package instance to scan
    :param attrs: None (return all package restrictions), or a sequence of
        specific attrs the package restriction must work against.
    """
    if not isinstance(restrict, (list, tuple)):
        restrict = [restrict]
    for r in restrict:
        if not isinstance(r, restriction.base):
            raise TypeError(
                'restrict must be of a restriction.base, '
                f'not {r.__class__.__class__}: {r!r}'
            )
    i = iflatten_func(restrict, _is_package_instance)
    if attrs is None:
        return i
    attrs = frozenset(attrs)
    return (
        r for r in i
        if invert == attrs.isdisjoint(getattr(r, 'attrs', ()))
    )



__all__ = ("skip_signatures",)

msg_header = "-----BEGIN PGP SIGNED MESSAGE-----\n"
msg_header_len = len(msg_header)
msg_hash = 'Hash:'
msg_hash_len = len(msg_hash)
sig_header = "-----BEGIN PGP SIGNATURE-----\n"
sig_header_len = len(sig_header)
sig_footer = "-----END PGP SIGNATURE-----\n"
sig_footer_len = len(sig_footer)

def skip_signatures(iterable):
    i = iter(iterable)
# format is-
#"""
#-----BEGIN PGP SIGNED MESSAGE-----
#Hash: SHA1
#
#"""

    for line in i:
        # so... prune msg first, then
        if line.endswith(msg_header):
            line = next(i)
            while line[msg_hash_len:] == msg_hash:
                line = next(i)
            # skip blank line after msg.
            next(i)
            continue
        while line.endswith(sig_header):
            line = next(i)
            # swallow the footer.
            while not line.endswith(sig_footer):
                line = next(i)
            # leave the next line on the stack
            line = next(i)

        yield line



from pkgcore.config import basics
from pkgcore.config.hint import ConfigHint, configurable
from pkgcore.ebuild import atom
from pkgcore.repository import util
from pkgcore.scripts import pquery
from pkgcore.test.scripts.helpers import ArgParseMixin
from snakeoil.test import TestCase


class FakeDomain:

    pkgcore_config_type = ConfigHint({'repos': 'refs:repo',
                                      'vdb': 'refs:repo'},
                                     typename='domain')

    def __init__(self, repos, vdb):
        object.__init__(self)
        self.source_repos = repos
        self.installed_repos = vdb


@configurable(typename='repo')
def fake_repo():
    return util.SimpleTree({'spork': {'foon': ('1', '2')}})


@configurable(typename='repo')
def fake_vdb():
    return util.SimpleTree({})


domain_config = basics.HardCodedConfigSection({
        'class': FakeDomain,
        'repos': [basics.HardCodedConfigSection({'class': fake_repo})],
        'vdb': [basics.HardCodedConfigSection({'class': fake_vdb})],
        'default': True,
        })


class CommandlineTest(TestCase, ArgParseMixin):

    _argparser = pquery.argparser

    def test_parser(self):
        self.assertError(
            'argument --min: not allowed with argument --max',
            '--max', '--min')
        self.parse('--all', domain=domain_config)

    def test_no_domain(self):
        self.assertError(
            "config error: no default object of type 'domain' found.  "
            "Please either fix your configuration, or set the domain via the --domain option.",
            '--all')

    def test_no_description(self):
        self.assertOut(
            [' * spork/foon-2',
             '     repo: MISSING',
             '     description: MISSING',
             '     homepage: MISSING',
             '     license: MISSING',
             '',
             ],
            '-v', '--max', '--all',
            test_domain=domain_config)

    def test_atom(self):
        config = self.parse(
            '--print-revdep', 'a/spork', '--all', domain=domain_config)
        self.assertEqual([atom.atom('a/spork')], config.print_revdep)

    def test_no_contents(self):
        self.assertOut([], '--contents', '--all', test_domain=domain_config)



import os
import re
import subprocess
import sys
from collections import defaultdict
from functools import partial
from weakref import WeakValueDictionary

from snakeoil import klass
from snakeoil.demandload import demand_compile_regexp
from snakeoil.mappings import ImmutableDict, OrderedFrozenSet, inject_getitem_as_getattr
from snakeoil.osutils import pjoin
from snakeoil.process.spawn import bash_version

from ..log import logger
from . import atom, const

demand_compile_regexp(
    '_valid_EAPI_regex', r"^[A-Za-z0-9_][A-Za-z0-9+_.-]*$"
)

eapi_optionals = ImmutableDict({
    # Controls whether PROPERTIES and RESTRICT are accumulated across eclasses.
    "accumulate_properties_restrict": False,

    # Controls what version of bash compatibility to force; see PMS.
    "bash_compat": '3.2',

    # Controls whether -r is allowed for dodoc.
    "dodoc_allow_recursive": False,

    # Controls the language awareness of doman; see PMS.
    "doman_language_detect": False,

    # Controls whether -i18n option is allowed.
    "doman_language_override": False,

    # Controls whether dosym -r option is allowed.
    "dosym_relative": False,

    # Controls whether an ebuild_phase function exists for ebuild consumption.
    'ebuild_phase_func': False,

    # Controls whether REPLACING vars are exported to ebuilds; see PMS.
    "exports_replacing": False,

    # Controls of whether failglob is enabled globally; see PMS.
    "global_failglob": False,

    # Controls whether MERGE vars are exported to ebuilds; see PMS.
    "has_merge_type": False,

    # Controls whether PORTDIR and ECLASSDIR are exported to ebuilds; see PMS.
    "has_portdir": True,

    # Controls whether DESTTREE and INSDESTTREE are exported during src_install; see PMS.
    "has_desttree": True,

    # Controls whether ROOT, EROOT, D, and ED end with a trailing slash; see PMS.
    "trailing_slash": os.sep,

    # Controls whether SYSROOT, ESYSROOT, and BROOT are defined; see PMS.
    "has_sysroot": False,

    # Controls whether package.provided files in profiles are supported; see PMS.
    "profile_pkg_provided": True,

    # Controls whether package.mask and other files in profiles can
    # be directories; see PMS.
    "has_profile_data_dirs": False,

    # Controls whether REQUIRED_USE is supported, enforcing constraints on
    # allowed use configuration states.
    "has_required_use": False,

    # Controls whether USE dependency defaults are supported, see PMS.
    "has_use_dep_defaults": False,

    # Controls whether ENV_UNSET is supported, see PMS.
    "has_env_unset": False,

    # Controls whether AA env var is exported to ebuilds; this is a flattened
    # listing of each filename in SRC_URI.
    "has_AA": True,

    # Controls whether KV (kernel version; see PMS for details) is exported.
    "has_KV": True,

    # Controls whether or not pkgcore, or extensions loaded, actually fully
    # support this EAPI.
    'is_supported': True,

    # Controls whether IUSE defaults are supported; see PMS.
    'iuse_defaults': False,

    # Controls whether new* style bash functions can take their content input
    # from stdin, rather than an explicit ondisk file.
    'new_reads_stdin': False,

    # Controls whether utilities die on failure; see PMS.
    'nonfatal': True,

    # Controls whether die supports a nonfatal option; see PMS.
    "nonfatal_die": False,

    # Controls whether this EAPI supports prefix related variables/settings;
    # prefix awareness basically. See PMS for full details.
    "prefix_capable": False,

    # Controls whether profile-defined IUSE injection is supported.
    "profile_iuse_injection": False,

    # Controls whether profiles support package.use.stable.* and use.stable.* files.
    "profile_stable_use": False,

    # Controls whether has_version/best_version supports --host-root option; see PMS.
    'query_host_root': False,

    # Controls whether has_version/best_version supports -b/-d/-r options; see PMS.
    'query_deps': False,

    # Controls whether SLOT values can actually be multi-part; see PMS EAPI 5.
    # This is related to ABI breakage detection.
    'sub_slotting': False,

    # Controls whether REQUIRED_USE supports the ?? operator.
    'required_use_one_of': False,

    # Controls whether SRC_URI supports the '->' operator for url filename renaming.
    "src_uri_renames": False,

    # Controls whether SRC_URI supports fetch+ and mirror+ prefixes.
    "src_uri_unrestrict": False,

    # Controls whether or not use dependency atoms are able to control their enforced
    # value relative to another; standard use deps just enforce either on or off; EAPIs
    # supporting this allow syntax that can enforce (for example) X to be on if Y is on.
    # See PMS EAPI 4 for full details.
    "transitive_use_atoms": False,

    # Controls whether or DEFINED_PHASES is mandated for this EAPI; if so, then we can
    # trust the cache definition and skip invoking those phases if they're not defined.
    # If the EAPI didn't mandate this var, then we can do our inference, but generally will
    # invoke the phase in the absense of that metadata var since we have no other choice.
    "trust_defined_phases_cache": False,

    # Controls whether unpack supports absolute paths; see PMS.
    "unpack_absolute_paths": False,

    # Controls whether unpack supports absolute paths; see PMS.
    "unpack_case_insensitive": False,

    # Regular expression to filter out valid update files
    "update_regex": re.compile(r'^([1-4])Q-(\d{4})$'),

    # Controls whether user patches are supported.
    "user_patches": False,
})


class _optionals_cls(ImmutableDict):

    inject_getitem_as_getattr(locals())


class EAPI(metaclass=klass.immutable_instance):

    known_eapis = WeakValueDictionary()
    unknown_eapis = WeakValueDictionary()

    def __init__(self, magic, parent=None, phases=(), default_phases=(),
                 mandatory_keys=(), dep_keys=(), metadata_keys=(),
                 eclass_keys=(), tracked_attributes=(), archive_exts=(),
                 optionals=None, ebd_env_options=None):
        sf = object.__setattr__

        sf(self, "_magic", str(magic))
        sf(self, "_parent", parent)

        sf(self, "phases", ImmutableDict(phases))
        sf(self, "phases_rev", ImmutableDict((v, k) for k, v in
           self.phases.items()))

        # We track the phases that have a default implementation- this is
        # primarily due to DEFINED_PHASES cache values not including it.
        sf(self, "default_phases", frozenset(default_phases))

        sf(self, "mandatory_keys", frozenset(mandatory_keys))
        sf(self, "dep_keys", frozenset(dep_keys))
        sf(self, "metadata_keys", (
            self.mandatory_keys | self.dep_keys | frozenset(metadata_keys)))
        # variables that eclasses have access to (used by pkgcheck eclass inherit checks)
        sf(self, "eclass_keys", self.mandatory_keys | self.dep_keys | frozenset(eclass_keys))
        sf(self, "tracked_attributes", (
            frozenset(tracked_attributes) | frozenset(x.lower() for x in dep_keys)))
        sf(self, "archive_exts", frozenset(archive_exts))

        if optionals is None:
            optionals = {}
        sf(self, 'options', _optionals_cls(optionals))
        if ebd_env_options is None:
            ebd_env_options = ()
        sf(self, "_ebd_env_options", ebd_env_options)

    @classmethod
    def register(cls, *args, **kwds):
        eapi = cls(*args, **kwds)
        pre_existing = cls.known_eapis.get(eapi._magic)
        if pre_existing is not None:
            raise ValueError(
                f"EAPI '{eapi}' is already known/instantiated- {pre_existing!r}")

        if (getattr(eapi.options, 'bash_compat', False) and
                bash_version() < eapi.options.bash_compat):
            # hard exit if the system doesn't have an adequate bash installed
            raise SystemExit(
                f"EAPI '{eapi}' requires >=bash-{eapi.options.bash_compat}, "
                f"system version: {bash_version()}")

        cls.known_eapis[eapi._magic] = eapi
        # generate EAPI bash libs when running from git repo
        eapi.bash_libs()
        return eapi

    @klass.jit_attr
    def is_supported(self):
        """Check if an EAPI is supported."""
        if EAPI.known_eapis.get(self._magic) is not None:
            if not self.options.is_supported:
                logger.warning(f"EAPI '{self}' isn't fully supported")
                sys.stderr.flush()
            return True
        return False

    @klass.jit_attr
    def bash_funcs_global(self):
        """Internally implemented global EAPI specific functions to skip when exporting."""
        # TODO: This is currently duplicated across EAPI objs, but
        # instead could be cached to a class attr.
        funcs = pjoin(const.EBD_PATH, '.generated', 'funcs', 'global')
        if not os.path.exists(funcs):
            # we're probably running in a cacheless git repo, so generate a cached version
            try:
                os.makedirs(os.path.dirname(funcs), exist_ok=True)
                with open(funcs, 'w') as f:
                    subprocess.run(
                        [pjoin(const.EBD_PATH, 'generate_global_func_list')],
                        cwd=const.EBD_PATH, stdout=f)
            except (IOError, subprocess.CalledProcessError) as e:
                raise Exception(
                    f"failed to generate list of global EAPI '{self}' specific functions: {str(e)}")

        with open(funcs, 'r') as f:
            return frozenset(line.strip() for line in f)

    @klass.jit_attr
    def bash_funcs(self):
        """Internally implemented EAPI specific functions to skip when exporting."""
        funcs = pjoin(const.EBD_PATH, '.generated', 'funcs', self._magic)
        if not os.path.exists(funcs):
            # we're probably running in a cacheless git repo, so generate a cached version
            try:
                os.makedirs(os.path.dirname(funcs), exist_ok=True)
                with open(funcs, 'w') as f:
                    subprocess.run(
                        [pjoin(const.EBD_PATH, 'generate_eapi_func_list'), self._magic],
                        cwd=const.EBD_PATH, stdout=f)
            except (IOError, subprocess.CalledProcessError) as e:
                raise Exception(
                    f"failed to generate list of EAPI '{self}' specific functions: {str(e)}")

        with open(funcs, 'r') as f:
            return frozenset(line.strip() for line in f)

    @klass.jit_attr
    def bash_cmds_internal(self):
        """EAPI specific commands for this EAPI."""
        cmds = pjoin(const.EBD_PATH, '.generated', 'cmds', self._magic, 'internal')
        if not os.path.exists(cmds):
            # we're probably running in a cacheless git repo, so generate a cached version
            try:
                os.makedirs(os.path.dirname(cmds), exist_ok=True)
                with open(cmds, 'w') as f:
                    subprocess.run(
                        [pjoin(const.EBD_PATH, 'generate_eapi_cmd_list'), '-i', self._magic],
                        cwd=const.EBD_PATH, stdout=f)
            except (IOError, subprocess.CalledProcessError) as e:
                raise Exception(
                    f'failed to generate list of EAPI {self} internal commands: {str(e)}')

        with open(cmds, 'r') as f:
            return frozenset(line.strip() for line in f)

    @klass.jit_attr
    def bash_cmds_deprecated(self):
        """EAPI specific commands deprecated for this EAPI."""
        cmds = pjoin(const.EBD_PATH, '.generated', 'cmds', self._magic, 'deprecated')
        if not os.path.exists(cmds):
            # we're probably running in a cacheless git repo, so generate a cached version
            try:
                os.makedirs(os.path.dirname(cmds), exist_ok=True)
                with open(cmds, 'w') as f:
                    subprocess.run(
                        [pjoin(const.EBD_PATH, 'generate_eapi_cmd_list'), '-d', self._magic],
                        cwd=const.EBD_PATH, stdout=f)
            except (IOError, subprocess.CalledProcessError) as e:
                raise Exception(
                    f'failed to generate list of EAPI {self} deprecated commands: {str(e)}')

        with open(cmds, 'r') as f:
            return frozenset(line.strip() for line in f)

    @klass.jit_attr
    def bash_cmds_banned(self):
        """EAPI specific commands banned for this EAPI."""
        cmds = pjoin(const.EBD_PATH, '.generated', 'cmds', self._magic, 'banned')
        if not os.path.exists(cmds):
            # we're probably running in a cacheless git repo, so generate a cached version
            try:
                os.makedirs(os.path.dirname(cmds), exist_ok=True)
                with open(cmds, 'w') as f:
                    subprocess.run(
                        [pjoin(const.EBD_PATH, 'generate_eapi_cmd_list'), '-b', self._magic],
                        cwd=const.EBD_PATH, stdout=f)
            except (IOError, subprocess.CalledProcessError) as e:
                raise Exception(
                    f'failed to generate list of EAPI {self} banned commands: {str(e)}')

        with open(cmds, 'r') as f:
            return frozenset(line.strip() for line in f)

    def bash_libs(self):
        """Generate internally implemented EAPI specific bash libs required by the ebd."""
        eapi_global_lib = pjoin(const.EBD_PATH, '.generated', 'libs', self._magic, 'global')
        script = pjoin(const.EBD_PATH, 'generate_eapi_lib')
        # skip generation when installing as the install process takes care of it
        if not os.path.exists(script):
            return

        if not os.path.exists(eapi_global_lib):
            try:
                os.makedirs(os.path.dirname(eapi_global_lib), exist_ok=True)
                with open(eapi_global_lib, 'w') as f:
                    subprocess.run(
                        [script, '-s', 'global', self._magic],
                        cwd=const.EBD_PATH, stdout=f)
            except (IOError, subprocess.CalledProcessError) as e:
                raise Exception(
                    f"failed to generate EAPI '{self}' global lib: {str(e)}")

        for phase in self.phases.values():
            eapi_lib = pjoin(const.EBD_PATH, '.generated', 'libs', self._magic, phase)
            if not os.path.exists(eapi_lib):
                try:
                    os.makedirs(os.path.dirname(eapi_lib), exist_ok=True)
                    with open(eapi_lib, 'w') as f:
                        subprocess.run(
                            [script, '-s', phase, self._magic],
                            cwd=const.EBD_PATH, stdout=f)
                except (IOError, subprocess.CalledProcessError) as e:
                    raise Exception(f"failed to generate EAPI '{self}' phase {phase} lib: {str(e)}")

    @klass.jit_attr
    def archive_exts_regex_pattern(self):
        """Regex pattern for supported archive extensions."""
        pattern = '|'.join(map(re.escape, self.archive_exts))
        if self.options.unpack_case_insensitive:
            return f'(?i:({pattern}))'
        return f'({pattern})'

    @klass.jit_attr
    def archive_exts_regex(self):
        """Regex matching strings ending with supported archive extensions."""
        return re.compile(rf'{self.archive_exts_regex_pattern}$')

    @klass.jit_attr
    def valid_slot_regex(self):
        """Regex matching valid SLOT values."""
        valid_slot = r'[A-Za-z0-9_][A-Za-z0-9+_.-]*'
        if self.options.sub_slotting:
            valid_slot += rf'(/{valid_slot})?'
        return re.compile(rf'^{valid_slot}$')

    @klass.jit_attr
    def atom_kls(self):
        return partial(atom.atom, eapi=self._magic)

    def interpret_cache_defined_phases(self, sequence):
        phases = set(sequence)
        if not self.options.trust_defined_phases_cache:
            if not phases:
                # run them all; cache was generated
                # by a pm that didn't support DEFINED_PHASES
                return frozenset(self.phases)

        phases.discard("-")
        return frozenset(phases)

    def __str__(self):
        return self._magic

    @klass.jit_attr
    def inherits(self):
        """Ordered set containing an EAPI's inheritance tree.

        Note that this assumes a simple, linear inheritance tree.
        """
        eapis = [self]
        eapi = self
        while eapi := eapi._parent:
            eapis.append(eapi)
        return OrderedFrozenSet(eapis)

    @klass.jit_attr
    def helpers(self):
        """Phase to directory mapping for EAPI specific helpers to add to $PATH."""
        paths = defaultdict(list)
        for eapi in self.inherits:
            paths['global'].append(pjoin(const.EBUILD_HELPERS_PATH, 'common'))
            helper_dir = pjoin(const.EBUILD_HELPERS_PATH, eapi._magic)
            for dirpath, dirnames, filenames in os.walk(helper_dir):
                if not filenames:
                    continue
                if dirpath == helper_dir:
                    paths['global'].append(dirpath)
                else:
                    phase = os.path.basename(dirpath)
                    if phase in self.phases_rev:
                        paths[phase].append(dirpath)
                    else:
                        raise ValueError(f'unknown phase: {phase!r}')
        return ImmutableDict((k, tuple(v)) for k, v in paths.items())

    @klass.jit_attr
    def ebd_env(self):
        """Dictionary of EAPI options passed to the ebd environment."""
        d = {}
        for k in self._ebd_env_options:
            d[f"PKGCORE_{k.upper()}"] = str(getattr(self.options, k)).lower()
        d["PKGCORE_EAPI_INHERITS"] = ' '.join(x._magic for x in self.inherits)
        d["EAPI"] = self._magic
        return ImmutableDict(d)


def get_eapi(magic, suppress_unsupported=True):
    """Return EAPI object for a given identifier."""
    if _valid_EAPI_regex.match(magic) is None:
        eapi_str = f" {magic!r}" if magic else ''
        raise ValueError(f'invalid EAPI{eapi_str}')
    eapi = EAPI.known_eapis.get(magic)
    if eapi is None and suppress_unsupported:
        eapi = EAPI.unknown_eapis.get(magic)
        if eapi is None:
            eapi = EAPI(magic=magic, optionals={'is_supported': False})
            EAPI.unknown_eapis[eapi._magic] = eapi
    return eapi


def _shorten_phase_name(func_name):
    if func_name.startswith(('src_', 'pkg_')):
        return func_name[4:]
    return func_name


def _mk_phase_func_map(*sequence):
    return {_shorten_phase_name(x): x for x in sequence}


def _combine_dicts(*mappings):
    return {k: v for d in mappings for k, v in d.items()}


# Note that pkg_setup is forced by default since this is how our env setup occurs.
common_default_phases = tuple(
    _shorten_phase_name(x) for x in
    ("pkg_setup", "src_unpack", "src_compile", "src_test", "pkg_nofetch"))

common_phases = (
    "pkg_setup", "pkg_config", "pkg_info", "pkg_nofetch",
    "pkg_prerm", "pkg_postrm", "pkg_preinst", "pkg_postinst",
    "src_unpack", "src_compile", "src_test", "src_install",
)

common_mandatory_metadata_keys = (
    "DESCRIPTION", "HOMEPAGE", "IUSE",
    "KEYWORDS", "LICENSE", "SLOT", "SRC_URI",
)

common_dep_keys = (
    "DEPEND", "RDEPEND", "PDEPEND",
)

common_metadata_keys = (
    "RESTRICT", "PROPERTIES", "DEFINED_PHASES", "INHERIT", "INHERITED", "EAPI",
)

common_eclass_keys = ("S", "RESTRICT", "PROPERTIES", "ECONF_SOURCE")

common_tracked_attributes = (
    "cflags", "cbuild", "chost", "ctarget", "cxxflags", "defined_phases",
    "description", "eapi", "distfiles", "fullslot", "homepage", "inherited",
    "iuse", "keywords", "ldflags", "license", "properties",
    "restrict", "source_repository",
)

common_archive_exts = (
    ".tar",
    ".tar.gz", ".tgz", ".tar.Z", ".tar.z",
    ".tar.bz2", ".tbz2", ".tbz",
    ".zip", ".ZIP", ".jar",
    ".gz", ".Z", ".z",
    ".bz2",
    ".rar", ".RAR",
    ".lha", ".LHa", ".LHA", ".lzh",
    ".a", ".deb",
    ".tar.lzma",
    ".lzma",
    ".7z", ".7Z",
)

# Boolean variables exported to the bash side, e.g. ebuild_phase_func is
# exported as PKGCORE_EBUILD_PHASE_FUNC.
common_env_optionals = (
    "bash_compat", "ebuild_phase_func", "global_failglob",
    "new_reads_stdin", "nonfatal", "nonfatal_die",
    "has_desttree", "accumulate_properties_restrict",
)

eapi0 = EAPI.register(
    magic="0",
    parent=None,
    phases=_mk_phase_func_map(*common_phases),
    default_phases=_mk_phase_func_map(*common_default_phases),
    mandatory_keys=common_mandatory_metadata_keys,
    dep_keys=common_dep_keys,
    metadata_keys=common_metadata_keys,
    eclass_keys=common_eclass_keys,
    tracked_attributes=common_tracked_attributes,
    archive_exts=common_archive_exts,
    optionals=eapi_optionals,
    ebd_env_options=common_env_optionals,
)

eapi1 = EAPI.register(
    magic="1",
    parent=eapi0,
    phases=eapi0.phases,
    default_phases=eapi0.default_phases,
    mandatory_keys=eapi0.mandatory_keys,
    dep_keys=eapi0.dep_keys,
    metadata_keys=eapi0.metadata_keys,
    eclass_keys=eapi0.eclass_keys,
    tracked_attributes=eapi0.tracked_attributes,
    archive_exts=eapi0.archive_exts,
    optionals=_combine_dicts(eapi0.options, dict(
        iuse_defaults=True,
    )),
    ebd_env_options=eapi0._ebd_env_options,
)

eapi2 = EAPI.register(
    magic="2",
    parent=eapi1,
    phases=_combine_dicts(
        eapi1.phases, _mk_phase_func_map("src_prepare", "src_configure")),
    default_phases=eapi1.default_phases.union(
        list(map(_shorten_phase_name, ["src_prepare", "src_configure"]))),
    mandatory_keys=eapi1.mandatory_keys,
    dep_keys=eapi1.dep_keys,
    metadata_keys=eapi1.metadata_keys,
    eclass_keys=eapi1.eclass_keys,
    tracked_attributes=eapi1.tracked_attributes,
    archive_exts=eapi1.archive_exts,
    optionals=_combine_dicts(eapi1.options, dict(
        doman_language_detect=True,
        transitive_use_atoms=True,
        src_uri_renames=True,
    )),
    ebd_env_options=eapi1._ebd_env_options,
)

eapi3 = EAPI.register(
    magic="3",
    parent=eapi2,
    phases=eapi2.phases,
    default_phases=eapi2.default_phases,
    mandatory_keys=eapi2.mandatory_keys,
    dep_keys=eapi2.dep_keys,
    metadata_keys=eapi2.metadata_keys,
    eclass_keys=eapi2.eclass_keys | frozenset(['EPREFIX', 'ED', 'EROOT']),
    tracked_attributes=eapi2.tracked_attributes,
    archive_exts=eapi2.archive_exts | frozenset([".tar.xz", ".xz"]),
    optionals=_combine_dicts(eapi2.options, dict(
        prefix_capable=True,
    )),
    ebd_env_options=eapi2._ebd_env_options,
)

eapi4 = EAPI.register(
    magic="4",
    parent=eapi3,
    phases=_combine_dicts(eapi3.phases, _mk_phase_func_map("pkg_pretend")),
    default_phases=eapi3.default_phases.union([_shorten_phase_name('src_install')]),
    mandatory_keys=eapi3.mandatory_keys,
    dep_keys=eapi3.dep_keys,
    metadata_keys=eapi3.metadata_keys | frozenset(["REQUIRED_USE"]),
    eclass_keys=eapi3.eclass_keys | frozenset(["DOCS", "REQUIRED_USE"]),
    tracked_attributes=eapi3.tracked_attributes,
    archive_exts=eapi3.archive_exts,
    optionals=_combine_dicts(eapi3.options, dict(
        dodoc_allow_recursive=True,
        doman_language_override=True,
        nonfatal=False,
        exports_replacing=True,
        has_AA=False, has_KV=False,
        has_merge_type=True,
        has_required_use=True,
        has_use_dep_defaults=True,
        trust_defined_phases_cache=True,
    )),
    ebd_env_options=eapi3._ebd_env_options,
)

eapi5 = EAPI.register(
    magic="5",
    parent=eapi4,
    phases=eapi4.phases,
    default_phases=eapi4.default_phases,
    mandatory_keys=eapi4.mandatory_keys,
    dep_keys=eapi4.dep_keys,
    metadata_keys=eapi4.metadata_keys,
    eclass_keys=eapi4.eclass_keys,
    tracked_attributes=eapi4.tracked_attributes | frozenset(["iuse_effective"]),
    archive_exts=eapi4.archive_exts,
    optionals=_combine_dicts(eapi4.options, dict(
        ebuild_phase_func=True,
        profile_iuse_injection=True,
        profile_stable_use=True,
        query_host_root=True,
        new_reads_stdin=True,
        required_use_one_of=True,
        sub_slotting=True,
    )),
    ebd_env_options=eapi4._ebd_env_options,
)

eapi6 = EAPI.register(
    magic="6",
    parent=eapi5,
    phases=eapi5.phases,
    default_phases=eapi5.default_phases,
    mandatory_keys=eapi5.mandatory_keys,
    dep_keys=eapi5.dep_keys,
    metadata_keys=eapi5.metadata_keys,
    eclass_keys=eapi5.eclass_keys | frozenset(["HTML_DOCS", "PATCHES"]),
    tracked_attributes=eapi5.tracked_attributes | frozenset(["user_patches"]),
    archive_exts=eapi5.archive_exts | frozenset([".txz"]),
    optionals=_combine_dicts(eapi5.options, dict(
        global_failglob=True,
        nonfatal_die=True,
        unpack_absolute_paths=True,
        unpack_case_insensitive=True,
        user_patches=True,
        bash_compat='4.2',
    )),
    ebd_env_options=eapi5._ebd_env_options,
)

eapi7 = EAPI.register(
    magic="7",
    parent=eapi6,
    phases=eapi6.phases,
    default_phases=eapi6.default_phases,
    mandatory_keys=eapi6.mandatory_keys,
    dep_keys=eapi6.dep_keys | frozenset(["BDEPEND"]),
    metadata_keys=eapi6.metadata_keys,
    eclass_keys=eapi6.eclass_keys,
    tracked_attributes=eapi6.tracked_attributes,
    archive_exts=eapi6.archive_exts,
    optionals=_combine_dicts(eapi6.options, dict(
        has_profile_data_dirs=True,
        has_portdir=False,
        has_desttree=False,
        profile_pkg_provided=False,
        query_host_root=False,
        query_deps=True,
        has_sysroot=True,
        has_env_unset=True,
        trailing_slash='',
    )),
    ebd_env_options=eapi6._ebd_env_options,
)

eapi8 = EAPI.register(
    magic="8",
    parent=eapi7,
    phases=eapi7.phases,
    default_phases=eapi7.default_phases,
    mandatory_keys=eapi7.mandatory_keys,
    dep_keys=eapi7.dep_keys | frozenset(["IDEPEND"]),
    metadata_keys=eapi7.metadata_keys,
    eclass_keys=eapi7.eclass_keys,
    tracked_attributes=eapi7.tracked_attributes,
    archive_exts=eapi7.archive_exts - frozenset([
        ".rar", ".RAR",
        ".lha", ".LHa", ".LHA", ".lzh",
        ".7z", ".7Z",
    ]),
    optionals=_combine_dicts(eapi7.options, dict(
        accumulate_properties_restrict=True,
        bash_compat='5.0',
        dosym_relative=True,
        src_uri_unrestrict=True,
        update_regex=re.compile(r'^[^.]'),
    )),
    ebd_env_options=eapi7._ebd_env_options,
)



"""Filter a bash environment dump."""

__all__ = ("run",)

import io
import re

from ..log import logger

COMMAND_PARSING, SPACE_PARSING = list(range(2))


def run(out, file_buff, var_match, func_match,
               global_envvar_callback=None,
               func_callback=None):
    """Print a filtered environment.

    :param out: file-like object to write to.
    :param file_buff: string containing the environment to filter.
        Should end in '\0'.
    :param var_match: result of build_regex_string or C{None}, for variables.
    :param func_match: result of build_regex_string or C{None}, for functions.
    """

    process_scope(out, file_buff, 0, var_match, func_match, '\0',
        global_envvar_callback, func_callback=func_callback)


def build_regex_string(tokens, invert=False):
    tokens = [_f for _f in tokens if _f]
    if not tokens:
        return None
    if len(tokens) == 1:
        s = tokens[0]
    else:
        s = f"(?:{'|'.join(tokens)})"
    s = f'^{s}$'
    if invert:
        s = f"(?!{s})"
    try:
        return re.compile(s)
    except re.error as e:
        raise Exception(f"failed compiling {s!r}:\n\nerror: {e}")


FUNC_LEN = len('function')

def is_function(buff, pos):
    """:return: start, end, pos or None, None, None tuple."""
    isspace = str.isspace
    try:
        while buff[pos] in ' \t':
            pos += 1
        if buff[pos:pos + FUNC_LEN] == 'function':
            try:
                if isspace(buff[pos + FUNC_LEN]):
                    pos += FUNC_LEN + 1
            except IndexError:
                # insane, but it could still be a function-
                # len('f(){:;}') <= FUNC_LEN.
                pass
        while isspace(buff[pos]):
            pos += 1
        start = pos
        while buff[pos] not in '\0 \t\n="\'()':
            pos += 1
        end = pos
        if end == start:
            return None, None, None
        while buff[pos] in ' \t':
            pos += 1
        if buff[pos] != '(':
            return None, None, None
        pos += 1
        while buff[pos] in ' \t':
            pos += 1
        if buff[pos] != ')':
            return None, None, None
        pos += 1
        while isspace(buff[pos]):
            pos += 1
        if buff[pos] != '{':
            return None, None, None
        return start, end, pos + 1
    except IndexError:
        # can't be a function, ran off the end
        return None, None, None


def is_envvar(buff, pos):
    """:return: start, end, pos or None, None, None tuple."""
    try:
        while buff[pos] in ' \t':
            pos += 1
        start = pos
        while True:
            if buff[pos] in '\0"\'()- \t\n':
                return None, None, None
            if buff[pos] == '=':
                if pos == start:
                    return None, None, None
                return start, pos, pos + 1
            pos += 1
    except IndexError:
        return None, None, None

def process_scope(out, buff, pos, var_match, func_match, endchar,
                  envvar_callback=None, func_callback=None,
                  func_level=0):
    window_start = pos
    window_end = None
    isspace = str.isspace
    end = len(buff)
    while pos < end and buff[pos] != endchar:
        # Wander forward to the next non space.
        if window_end is not None:
            if out is not None:
                out.write(buff[window_start:window_end].encode('utf-8'))
            window_start = pos
            window_end = None
        com_start = pos
        ch = buff[pos]
        if isspace(ch):
            pos += 1
            continue

        # Ignore comments.
        if ch == '#':
            pos = walk_statement_pound(buff, pos, endchar)
            continue

        new_start, new_end, new_p = is_function(buff, pos)
        if new_p is not None:
            func_name = buff[new_start:new_end]
            logger.debug(f'matched func name {func_name!r}')
            new_p = process_scope(None, buff, new_p, None, None, '}',
                                  func_callback=func_callback,
                                  func_level=func_level+1)
            logger.debug(f'ended processing {func_name!r}')
            if func_callback is not None:
                func_callback(func_level, func_name, buff[new_start:new_p])
            if func_match is not None and func_match(func_name):
                logger.debug(f'filtering func {func_name!r}')
                window_end = com_start
            pos = new_p
            pos += 1
            continue
        # Check for env assignment.
        new_start, new_end, new_p = is_envvar(buff, pos)
        if new_p is None:
            # Non env assignment.
            pos = walk_command_complex(buff, pos, endchar, COMMAND_PARSING)
            # icky icky icky icky
            if pos < end and buff[pos] != endchar:
                pos += 1
        else:
            # Env assignment.
            var_name = buff[new_start:new_end]
            pos = new_p
            if envvar_callback:
                envvar_callback(var_name)
            logger.debug(f'matched env assign {var_name!r}')

            if var_match is not None and var_match(var_name):
                # This would be filtered.
                logger.info(f"filtering var {var_name!r}")
                window_end = com_start

            if pos >= end:
                return pos

            while (pos < end and not isspace(buff[pos])
                   and buff[pos] != ';'):
                if buff[pos] == "'":
                    pos = walk_statement_no_parsing(buff, pos + 1, "'") + 1
                elif buff[pos] in '"`':
                    pos = walk_command_escaped_parsing(buff, pos + 1,
                                                       buff[pos]) + 1
                elif buff[pos] == '(':
                    pos = walk_command_escaped_parsing(buff, pos + 1, ')') + 1
                elif buff[pos] == '$':
                    pos += 1
                    if pos >= end:
                        continue
                    pos = walk_dollar_expansion(buff, pos, end, endchar)
                    continue
                else:
                    # blah=cah ; single word
                    pos = walk_command_complex(buff, pos, ' ', SPACE_PARSING)

    if out is not None:
        if window_end is None:
            window_end = pos
        if window_end > end:
            window_end = end
        out.write(buff[window_start:window_end].encode('utf-8'))

    return pos


def walk_statement_no_parsing(buff, pos, endchar):
    pos = buff.find(endchar, pos)
    if pos == -1:
        pos = len(buff) - 1
    return pos


def walk_statement_dollared_quote_parsing(buff, pos, endchar):
    end = len(buff)
    while pos < end:
        if buff[pos] == endchar:
            return pos
        elif buff[pos] == '\\':
            pos += 1
        pos += 1
    return pos


def walk_here_statement(buff, pos):
    pos += 1
    logger.debug('starting here processing for COMMAND for level 2 at p == %.10s', pos)
    if buff[pos] == '<':
        logger.debug("correction, it's a third level here. Handing back to command parsing")
        return pos + 1
    isspace = str.isspace
    end = len(buff)
    while pos < end and (isspace(buff[pos]) or buff[pos] == '-'):
        pos += 1
    if buff[pos] in "'\"":
        end_here = walk_statement_no_parsing(buff, pos + 1, buff[pos])
        pos += 1
    else:
        end_here = walk_command_complex(buff, pos, ' ', SPACE_PARSING)
    here_word = buff[pos:end_here]
    logger.debug(f'matched len({len(here_word)})/{here_word!r} for a here word')
    # XXX watch this. Potential for horkage. Need to do the quote
    # removal thing. This sucks.
    end_here += 1
    if end_here >= end:
        return end_here

    here_len = len(here_word)
    end_here = buff.find(here_word, end_here)
    while end_here != -1:
        i = here_len + end_here
        if buff[i] in ';\n\r})':
            i = end_here - 1
            while i >= 0 and buff[i] in '\t ':
                i -= 1
            if i >= 0 and buff[i] == '\n':
                break
        end_here = buff.find(here_word, end_here + here_len)

    if end_here == -1:
        return end
    return end_here + len(here_word)


def walk_statement_pound(buff, pos, endchar=None):
    if pos and not buff[pos-1].isspace():
        return pos + 1
    if endchar == '`':
        i = buff.find('\n', pos)
        i2 = buff.find(endchar, pos)
        if i == -1:
            if i2 != -1:
                return i2
        else:
            if i2 != -1:
                return min(i, i2)
            return i
        return len(buff) - 1

    pos = buff.find('\n', pos)
    if pos == -1:
        pos = len(buff) - 1
    return pos


def walk_command_complex(buff, pos, endchar, interpret_level):
    start = pos
    isspace = str.isspace
    end = len(buff)
    while pos < end:
        ch = buff[pos]
        if ch == endchar:
            if endchar != '}':
                return pos
            if start == pos:
                return pos
            if buff[pos - 1] in ";\n":
                return pos
        elif (interpret_level == COMMAND_PARSING and ch in ';\n') or \
            (interpret_level == SPACE_PARSING and isspace(ch)):
            return pos
        elif ch == '\\':
            pos += 1
        elif ch == '<':
            if (pos < end - 1 and buff[pos + 1] == '<' and
                interpret_level == COMMAND_PARSING):
                pos = walk_here_statement(buff, pos + 1)
                # we continue immediately; walk_here deposits us at the end
                # of the here op, not consuming the final delimiting char
                # since it may be an endchar
                continue
            else:
                logger.debug(f'noticed <, interpret_level={interpret_level}')
        elif ch == '#':
            if start == pos or isspace(buff[pos - 1]) or buff[pos - 1] == ';':
                pos = walk_statement_pound(buff, pos)
                continue
        elif ch == '$':
            pos = walk_dollar_expansion(buff, pos + 1, end, endchar)
            continue
        elif ch == '{':
            pos = walk_command_escaped_parsing(buff, pos + 1, '}')
        elif ch == '(' and interpret_level == COMMAND_PARSING:
            pos = walk_command_escaped_parsing(buff, pos + 1, ')')
        elif ch in '`"':
            pos = walk_command_escaped_parsing(buff, pos + 1, ch)
        elif ch == "'" and endchar != '"':
            pos = walk_statement_no_parsing(buff, pos +1, "'")
        pos += 1
    return pos

def raw_walk_command_escaped_parsing(buff, pos, endchar):
    end = len(buff)
    while pos < end:
        ch = buff[pos]
        if ch == endchar:
            return pos
        elif ch == '\\':
            pos += 1
        elif ch == '{':
            if endchar != '"':
                pos = raw_walk_command_escaped_parsing(
                    buff, pos + 1, '}')
        elif ch == '(':
            if endchar != '"':
                pos = raw_walk_command_escaped_parsing(
                    buff, pos + 1, ')')
        elif ch in '`"':
            pos = raw_walk_command_escaped_parsing(buff, pos + 1, ch)
        elif ch == "'" and endchar != '"':
            pos = walk_statement_no_parsing(buff, pos + 1, "'")
        elif ch == '$':
            pos = walk_dollar_expansion(buff, pos + 1, end, endchar,
                disable_quote = endchar == '"')
            continue
        elif ch == '#' and endchar != '"':
            pos = walk_statement_pound(buff, pos, endchar)
            continue
        pos += 1
    return pos

walk_command_escaped_parsing = raw_walk_command_escaped_parsing

def walk_dollar_expansion(buff, pos, end, endchar, disable_quote=False):
    if buff[pos] == '(':
        return process_scope(None, buff, pos + 1, None, None, ')') + 1
    if buff[pos] == "'" and not disable_quote:
        return walk_statement_dollared_quote_parsing(buff, pos +1, "'") + 1
    if buff[pos] != '{':
        if buff[pos] == '$':
            # short circuit it.
            return pos + 1
        while pos < end and buff[pos] != endchar:
            if buff[pos].isspace():
                return pos
            if buff[pos] == '$':
                # shouldn't this be passing disable_quote ?
                return walk_dollar_expansion(buff, pos + 1, end, endchar)
            if not buff[pos].isalnum():
                if buff[pos] != '_':
                    return pos
            pos += 1

        if pos >= end:
            return end
        return pos

    pos += 1
    # shortcut ${$} to avoid going too deep. ${$a} isn't valid, so no concern
    if pos == '$':
        return pos + 1
    while pos < end and buff[pos] != '}':
        if buff[pos] == '$':
            # disable_quote?
            pos = walk_dollar_expansion(buff, pos + 1, end, endchar)
        else:
            pos += 1
    return pos + 1


def main_run(out_handle, data, vars_to_filter=(), funcs_to_filter=(), vars_is_whitelist=False, funcs_is_whitelist=False,
             global_envvar_callback=None, func_callback=None):
    vars = funcs = None
    if vars_to_filter:
        vars = build_regex_string(vars_to_filter, invert=vars_is_whitelist).match

    if funcs_to_filter:
        if isinstance(funcs_to_filter, str):
            raise ValueError("funcs_str should not be a string; should be a sequence.")
        funcs = build_regex_string(funcs_to_filter, invert=funcs_is_whitelist).match

    data = data + '\0'
    kwds = {'global_envvar_callback': global_envvar_callback}

    if func_callback:
        kwds['func_callback'] = func_callback
    if out_handle is None:
        out_handle = io.BytesIO()

    run(out_handle, data, vars, funcs, **kwds)



"""configuration subsystem"""

__all__ = ('load_config',)

# keep these imports as minimal as possible; access to
# pkgcore.config isn't uncommon, thus don't trigger till
# actually needed

import os

from .. import const
from . import central, cparser


def load_config(user_conf_file=const.USER_CONF_FILE,
                system_conf_file=const.SYSTEM_CONF_FILE,
                debug=False, prepend_sources=(),
                skip_config_files=False, profile_override=None,
                location=None, **kwargs):
    """The main entry point for any code looking to use pkgcore.

    Args:
        user_conf_file (optional[str]): pkgcore user config file path
        system_conf_file (optional[str]): system pkgcore config file path
        profile_override (optional[str]): targeted profile instead of system setting
        location (optional[str]): path to pkgcore config file or portage config directory
        skip_config_files (optional[str]): don't attempt to load any config files

    Returns:
        :obj:`pkgcore.config.central.ConfigManager` instance: system config
    """
    configs = list(prepend_sources)
    if not skip_config_files:
        # load a pkgcore config file if one exists
        for config in (location, user_conf_file, system_conf_file):
            if config is not None and os.path.isfile(config):
                with open(config) as f:
                    configs.append(cparser.config_from_file(f))
                break
        # otherwise load the portage config
        else:
            # delay importing to avoid circular imports
            from pkgcore.ebuild.portage_conf import PortageConfig
            configs.append(PortageConfig(
                location=location, profile_override=profile_override, **kwargs))
    return central.CompatConfigManager(central.ConfigManager(configs, debug=debug))



# potentially use an intermediate base for user config errors,
# separate base for instantiation?

"""Exceptions raised by the config code."""

__all__ = (
    "TypeDefinitionError", "ConfigurationError", "ParsingError",
    "CollapseInheritOnly", "ComplexInstantiationError",
    "QuoteInterpretationError",
)

from ..exceptions import PkgcoreException, PkgcoreUserException


def _identify_functor_source(functor):
    module = getattr(functor, '__module__', None)
    if module is None:
        return functor.__name__
    return f'{module}.{functor.__name__}'


class ConfigError(PkgcoreException):
    """Generic config exception."""


class UserConfigError(ConfigError, PkgcoreUserException):
    """Generic config exception with user relevant error."""


class TypeDefinitionError(ConfigError):
    """Fatal error in type construction."""


class ConfigurationError(ConfigError):
    """Fatal error in parsing a config section.

    :type stack: sequence of strings.
    :ivar stack: messages describing where this ConfigurationError originated.
        configuration-related code catching ConfigurationError that wants to
        raise its own ConfigurationError should modify (usually append to)
        the stack and then re-raise the original exception (this makes sure
        the traceback is preserved).
    """

    def __init__(self, message):
        super().__init__(message)
        self.stack = [message]

    def __str__(self):
        return ':\n'.join(reversed(self.stack))


class ParsingError(ConfigurationError, PkgcoreUserException):
    """Generic file parsing exception."""

    def __init__(self, message=None, exception=None):
        if message is not None:
            super().__init__(message)
        elif exception is not None:
            super().__init__(str(exception))
        else:
            raise ValueError('specify at least one of message and exception')
        self.message = message
        self.exc = exception

    def __str__(self):
        msg = f'parsing failed: {self.message}'
        if self.exc is not None:
            msg += f'\n{self.exc}'
        return msg


class CollapseInheritOnly(ConfigurationError):
    """Attempt was made to collapse an uncollapsable section.

    Separate exception because pconfig catches it separately.
    """


class InstantiationError(ConfigurationError):

    _txt = "Failed instantiating section %r%s"

    def __init__(self, section_name, message=None):
        self.section_name = section_name
        self.message = message

    def __str__(self):
        s = self.message
        if s is None:
            s = ''
        else:
            s = ': %s' % (s,)
        return self._txt % (self.section_name, s)


class AutoloadInstantiationError(InstantiationError):

    _txt = "Failed loading autoload section %r%s"


class ComplexInstantiationError(ConfigurationError):
    """Exception occurred during instantiation.

    :ivar callable: callable object which failed during instantiation.
    :ivar pargs: positional args passed to callable.
    :ivar kwargs: keyword args passed to callable.
    :ivar exc: original exception object or None.

    A well-behaved configurable callable should raise this exception
    if instantiation failed, providing one or both of message and
    exception. The other fields will be filled in by central.

    If the callable raises something else central will wrap it in
    this, but that will lose the traceback.
    """

    def __init__(self, message=None, exception=None, callable_obj=None,
                 pargs=None, kwargs=None):
        if message is not None:
            super().__init__(message)
        elif exception is not None:
            super().__init__(str(exception))
        else:
            raise ValueError('specify at least one of message and exception')
        self.message = message
        self.callable = callable_obj
        self.pargs = pargs
        self.kwargs = kwargs
        self.exc = exception

    def __str__(self):
        # self.callable should be set here (nothing should try to catch
        # and str() this before central had a chance to fill it in)
        if self.message is not None:
            if self.callable is None:
                message = f'{self.message!r}, callable unset!'
            else:
                message = (
                    f'{self.message!r} instantiating '
                    f'{self.callable.__module__}.{self.callable.__name__}'
                )
        # The weird repr(str(exc)) used here quotes the message nicely.
        elif self.callable is not None:
            message = (
                f'Caught exception {str(self.exc)!r} '
                f'instantiating {self.callable.__module__}.{self.callable.__name__}'
            )
        else:
            message = f'Caught exception {str(self.exc)!r}, callable unset!'
        return ':\n'.join(reversed([message] + self.stack[1:]))


class QuoteInterpretationError(ConfigurationError):
    """Quoting of a var was screwed up."""

    def __init__(self, string):
        super().__init__(f'parsing of {string!r} failed')
        self.str = string



import sys

import pytest
from pkgcore.config import basics, errors
from pkgcore.config.hint import configurable
from pkgcore.scripts import pconfig
from pkgcore.test.scripts.helpers import ArgParseMixin
from snakeoil.test import TestCase


@configurable({'reff': 'ref:spork'})
def spork(reff):
    """Test thing."""

def foon():
    pass

@configurable(typename='spork')
def pseudospork():
    pass

@configurable(typename='multi', allow_unknowns=True, types={
        'string': 'str', 'boolean': 'bool', 'list': 'list',
        'callable': 'callable', 'lazy_ref': 'lazy_ref:spork',
        'ref': 'ref:spork', 'lazy_refs': 'lazy_refs:spork',
        'refs': 'refs:spork',
        })
def multi(**kwargs):
    """Just something taking all kinds of args."""

# "in positional but not in required" is an error.
@configurable(positional=['foon'])
def broken_type(*args):
    """Noop."""

@configurable(types={'inc': 'list'}, allow_unknowns=True)
def increment(inc=()):
    """Noop."""


class DescribeClassTest(ArgParseMixin):

    _argparser = pconfig.describe_class

    def test_parser(self, capsys):
        self.assertError(
            'the following arguments are required: target_class')

        module = "module 'pkgcore'"

        with pytest.raises(SystemExit):
            self.parse('pkgcore.spork')
        captured = capsys.readouterr()
        assert captured.err.strip() == (
            f"argument target_class: Failed importing target 'pkgcore.spork': '{module} has no attribute 'spork''")

        with pytest.raises(SystemExit):
            self.parse('pkgcore.a', 'pkgcore.b')
        captured = capsys.readouterr()
        assert captured.err.strip() == (
            f"argument target_class: Failed importing target 'pkgcore.a': '{module} has no attribute 'a''")

        self.parse('pkgcore.scripts')

    def test_describe_class(self):
        self.assertOut(
            ['typename is spork',
             'Test thing.',
             '',
             'reff: ref:spork (required)'],
            'tests.scripts.test_pconfig.spork')
        self.assertOut(
            ['typename is increment',
             'Noop.',
             'values not listed are handled as strings',
             '',
             'inc: list'],
            'tests.scripts.test_pconfig.increment')

    def test_broken_type(self):
        self.assertErr(
            ['Not a valid type!'],
            'tests.scripts.test_pconfig.broken_type')


class ClassesTest(TestCase, ArgParseMixin):

    _argparser = pconfig.classes

    def test_classes(self):
        sections = []
        for i in range(10):
            @configurable(typename='spork')
            def noop():
                """noop"""
            noop.__name__ = str(i)
            sections.append(basics.HardCodedConfigSection({'class': noop}))
        self.assertOut(
            ['tests.scripts.test_pconfig.foon'],
            spork=basics.HardCodedConfigSection({'class': foon}))
        self.assertOut([
                'tests.scripts.test_pconfig.0',
                'tests.scripts.test_pconfig.1',
                'tests.scripts.test_pconfig.2',
                'tests.scripts.test_pconfig.3',
                'tests.scripts.test_pconfig.4',
                'tests.scripts.test_pconfig.5',
                'tests.scripts.test_pconfig.multi',
                'tests.scripts.test_pconfig.pseudospork',
                'tests.scripts.test_pconfig.spork',
                ],
            bork=basics.HardCodedConfigSection({
                    'class': pseudospork, 'bork': True, 'inherit-only': True}),
            multi=basics.HardCodedConfigSection({
                    'class': multi,
                    'ref': sections[0],
                    'refs': sections[1:3],
                    'lazy_ref': sections[3],
                    'lazy_refs': sections[4:6],
                    'random': 'unknown',
                    }),
            spork=basics.HardCodedConfigSection({
                    'class': spork,
                    'reff': basics.HardCodedConfigSection({
                            'class': pseudospork})}))


class DumpTest(TestCase, ArgParseMixin):

    _argparser = pconfig.dump

    def test_dump(self):
        self.assertOut(
            ["'spork' {",
             '    # typename of this section: foon',
             '    class tests.scripts.test_pconfig.foon;',
             '}',
             ''],
            spork=basics.HardCodedConfigSection({'class': foon}))

    def test_default(self):
        self.assertOut(
            ["'spork' {",
             '    # typename of this section: foon',
             '    class tests.scripts.test_pconfig.foon;',
             '    default true;',
             '}',
             ''],
            spork=basics.HardCodedConfigSection({'class': foon,
                                                 'default': True}))

    def test_uncollapsable(self):
        self.assertOut(
            '',
            spork=basics.HardCodedConfigSection({
                    'class': foon, 'broken': True, 'inherit-only': True}))

    def test_serialise(self):
        nest = basics.HardCodedConfigSection({'class': pseudospork})
        self.assertOut(
            ["'spork' {",
             '    # typename of this section: multi',
             '    class tests.scripts.test_pconfig.multi;',
             '    # type: bool',
             '    boolean True;',
             '    # type: callable',
             '    callable tests.scripts.test_pconfig.multi;',
             '    # type: lazy_ref:spork',
             '    lazy_ref {',
             '        # typename of this section: spork',
             '        class tests.scripts.test_pconfig.pseudospork;',
             '    };',
             '    # type: lazy_refs:spork',
             '    lazy_refs {',
             '        # typename of this section: spork',
             '        class tests.scripts.test_pconfig.pseudospork;',
             '    } {',
             '        # typename of this section: spork',
             '        class tests.scripts.test_pconfig.pseudospork;',
             '    };',
             '    # type: list',
             "    list 'a' 'b\\' \"c';",
             '    # type: ref:spork',
             '    ref {',
             '        # typename of this section: spork',
             '        class tests.scripts.test_pconfig.pseudospork;',
             '    };',
             '    # type: refs:spork',
             '    refs {',
             '        # typename of this section: spork',
             '        class tests.scripts.test_pconfig.pseudospork;',
             '    } {',
             '        # typename of this section: spork',
             '        class tests.scripts.test_pconfig.pseudospork;',
             '    };',
             '    # type: str',
             '    string \'it is a "stringy" \\\'string\\\'\';',
             '    # type: str',
             "    unknown 'random';",
             '}',
             ''],
            spork=basics.HardCodedConfigSection({
                    'class': multi,
                    'string': 'it is a "stringy" \'string\'',
                    'boolean': True,
                    'list': ['a', 'b\' "c'],
                    'callable': multi,
                    'ref': nest,
                    'lazy_ref': nest,
                    'refs': [nest, nest],
                    'lazy_refs': [nest, nest],
                    'unknown': 'random',
                    }))

    def test_one_typename(self):
        self.assertOut(
            ["'spork' {",
             '    # typename of this section: spork',
             '    class tests.scripts.test_pconfig.pseudospork;',
             '}',
             '',
             ],
            'spork',
            spork=basics.HardCodedConfigSection({'class': pseudospork}),
            foon=basics.HardCodedConfigSection({'class': foon}),
            )


class UncollapsableTest(TestCase, ArgParseMixin):

    _argparser = pconfig.uncollapsable

    def test_uncollapsable(self):
        self.assertOut(
            ["section foon:",
            " Collapsing section named 'foon'",
            " cannot collapse inherit-only section"
            "",
            "",
            "section spork:",
            " Collapsing section named 'spork'",
            " type tests.scripts.test_pconfig.spork needs settings for 'reff'"
            "",
            "",
            ],
            spork=basics.HardCodedConfigSection({'class': spork}),
            foon=basics.HardCodedConfigSection({'class': spork,
                                                'inherit-only': True}),
            )


class ConfigurablesTest(TestCase, ArgParseMixin):

    _argparser = pconfig.configurables

    def test_configurables(self):
        self.assertError(
            'unrecognized arguments: bar',
            'foo', 'bar')


class WeirdSection(basics.ConfigSection):

    def __contains__(self, key):
        return key == 'sects'

    def keys(self):
        return ['sects']

    def render_value(self, central, name, arg_type):
        if name != 'sects':
            raise KeyError(name)
        if arg_type != 'repr':
            raise errors.ConfigurationError(f'{arg_type!r} unsupported')
        return 'refs', [
            ['spork', basics.HardCodedConfigSection({'foo': 'bar'})],
            None, None]


class DumpUncollapsedTest(TestCase, ArgParseMixin):

    _argparser = pconfig.dump_uncollapsed

    def test_dump_uncollapsed(self):
        self.assertOut(
            ['# Warning:',
             '# Do not copy this output to a configuration file directly,',
             '# because the types you see here are only guesses.',
             '# A value used as "list" in the collapsed config will often',
             '# show up as "string" here and may need to be converted',
             '# (for example from space-separated to comma-separated)',
             '# to work in a config file with a different format.',
             '',
             '********',
             'Source 1',
             '********',
             '',
             'foon',
             '====',
             '# type: callable',
             "'class' = tests.scripts.test_pconfigspork",
             '# type: bool',
             "'inherit-only' = True",
             '# type: refs',
             "'refs' = ",
             '    nested section 1',
             '    ================',
             '    # type: str',
             "    'crystal' = 'clear'",
             '',
             '    nested section 2',
             '    ================',
             '    # type: refs',
             "    'sects.prepend' = ",
             '        nested section 1',
             '        ================',
             "        named section 'spork'",
             '',
             '        nested section 2',
             '        ================',
             '        # type: str',
             "        'foo' = 'bar'",
             '',
             '',
             '# type: list',
             "'seq' = 'a' 'b c'",
             '# type: str',
             "'str' = 'quote \\'\" unquote'",
             '',
             'spork',
             '=====',
             '# type: callable',
             "'class' = tests.scripts.test_pconfigspork",
             '',
             ],
            spork=basics.HardCodedConfigSection({'class': spork}),
            foon=basics.HardCodedConfigSection({
                    'class': spork,
                    'inherit-only': True,
                    'refs': [
                        basics.HardCodedConfigSection({'crystal': 'clear'}),
                        WeirdSection(),
                        ],
                    'seq': ['a', 'b c'],
                    'str': 'quote \'" unquote',
                    }),
            )



"""
exceptions thrown by the MergeEngine
"""

__all__ = ("ModificationError", "BlockModification", "TriggerUnknownCset")


from ..exceptions import PkgcoreException


class ModificationError(PkgcoreException):
    """Base Exception class for modification errors"""

    def __init__(self, trigger, msg):
        self.trigger = trigger
        self.msg = msg
        super().__init__(f"{self.trigger}: modification error: {self.msg}")


class BlockModification(ModificationError):
    """Merging cannot proceed"""

    def __str__(self):
        return "Modification was blocked by %s: %s" % (
            self.trigger.__class__.__name__, self.msg)

class TriggerUnknownCset(ModificationError):
    """Trigger's required content set isn't known"""

    def __init__(self, trigger, csets):
        if not isinstance(csets, (tuple, list)):
            csets = (csets,)
        super().__init__(
            f"{self.__class__}: trigger {trigger!r} unknown cset: {csets!r}")
        self.trigger, self.csets = trigger, csets
